import numpy as np
import pytest

from psyneulink.core.components.functions.function import FunctionError
from psyneulink.core.components.functions.learningfunctions import Hebbian, Reinforcement, TDLearning
from psyneulink.core.components.functions.objectivefunctions import Stability, Distance
from psyneulink.core.components.functions.distributionfunctions import NormalDist, ExponentialDist, \
    UniformDist, GammaDist, WaldDist, DriftDiffusionAnalytical
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import SimpleIntegrator, \
    AdaptiveIntegrator, DriftDiffusionIntegrator, OrnsteinUhlenbeckIntegrator, FitzHughNagumoIntegrator, \
    AccumulatorIntegrator, DualAdaptiveIntegrator
from psyneulink.core.components.functions.transferfunctions import Linear, Exponential, Logistic, SoftMax, LinearMatrix
from psyneulink.core.components.functions.combinationfunctions import Reduce, LinearCombination, CombineMeans
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.globals.keywords import \
    MAX_ABS_INDICATOR, MAX_ABS_ONE_HOT, MAX_ABS_VAL, MAX_INDICATOR, MAX_ONE_HOT, MAX_VAL, \
    MEAN, MEDIAN, PROB, STANDARD_DEVIATION, VARIANCE

class TestProcessingMechanismFunctions:

    def test_processing_mechanism_linear_function(self):

        PM1 = ProcessingMechanism()
        PM1.execute(1.0)
        assert np.allclose(PM1.value, 1.0)

        PM2 = ProcessingMechanism(function=Linear(slope=2.0,
                                                  intercept=1.0))
        PM2.execute(1.0)
        assert np.allclose(PM2.value, 3.0)

    def test_processing_mechanism_LinearCombination_function(self):

        PM1 = ProcessingMechanism(function=LinearCombination)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Reduce_function(self):
        PM1 = ProcessingMechanism(function=Reduce)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_CombineMeans_function(self):
        PM1 = ProcessingMechanism(function=CombineMeans)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Exponential_function(self):
        PM1 = ProcessingMechanism(function=Exponential)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Logistic_function(self):
        PM1 = ProcessingMechanism(function=Logistic)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_SoftMax_function(self):
        PM1 = ProcessingMechanism(function=SoftMax(per_item=False))
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_SimpleIntegrator_function(self):
        PM1 = ProcessingMechanism(function=SimpleIntegrator)
        PM1.execute(1.0)

    def test_processing_mechanism_AdaptiveIntegrator_function(self):
        PM1 = ProcessingMechanism(function=AdaptiveIntegrator)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_DriftDiffusionIntegrator_function(self):
        PM1 = ProcessingMechanism(function=DriftDiffusionIntegrator)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_OrnsteinUhlenbeckIntegrator_function(self):
        PM1 = ProcessingMechanism(function=OrnsteinUhlenbeckIntegrator)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_AccumulatorIntegrator_function(self):
        PM1 = ProcessingMechanism(function=AccumulatorIntegrator)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_FitzHughNagumoIntegrator_function(self):
        PM1 = ProcessingMechanism(function=FitzHughNagumoIntegrator)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_DualAdaptiveIntegrator_function(self):
        PM1 = ProcessingMechanism(function=DualAdaptiveIntegrator)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_BogaczEtAl_function(self):
        PM1 = ProcessingMechanism(function=DriftDiffusionAnalytical)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    # COMMENTED OUT BECAUSE OF MATLAB ENGINE:
    # def test_processing_mechanism_NavarroAndFuss_function(self):
    #     PM1 = ProcessingMechanism(function=NavarroAndFuss)
    #     PM1.execute(1.0)
    #     # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_NormalDist_function(self):
        PM1 = ProcessingMechanism(function=NormalDist)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_ExponentialDist_function(self):
        PM1 = ProcessingMechanism(function=ExponentialDist)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_UniformDist_function(self):
        PM1 = ProcessingMechanism(function=UniformDist)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_GammaDist_function(self):
        PM1 = ProcessingMechanism(function=GammaDist)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_WaldDist_function(self):
        PM1 = ProcessingMechanism(function=WaldDist)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Stability_function(self):
        PM1 = ProcessingMechanism(function=Stability)
        PM1.execute(1.0)
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Distance_function(self):
        PM1 = ProcessingMechanism(function=Distance,
                                  default_variable=[[0,0], [0,0]])
        PM1.execute([[1, 2], [3, 4]])
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Hebbian_function(self):
        PM1 = ProcessingMechanism(function=Hebbian,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Reinforcement_function(self):
        PM1 = ProcessingMechanism(function=Reinforcement,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # assert np.allclose(PM1.value, 1.0)

    # COMMENTING OUT BECAUSE BACK PROP FN DOES NOT WORK WITH UNRESTRICTED MECHANISM
    # def test_processing_mechanism_BackPropagation_function(self):
    #     PM1 = ProcessingMechanism(function=BackPropagation,
    #                                          default_variable=[[0.0], [0.0], [0.0]])
    #     PM1.execute([[1.0], [2.0], [3.0]])
    #     PM1.execute(1.0)
    #     # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_TDLearning_function(self):
        PM1 = ProcessingMechanism(function=TDLearning,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_multiple_input_ports(self):
        PM1 = ProcessingMechanism(size=[4, 4], function=LinearCombination, input_ports=['input_1', 'input_2'])
        PM2 = ProcessingMechanism(size=[2, 2, 2], function=LinearCombination, input_ports=['1', '2', '3'])
        PM1.execute([[1, 2, 3, 4], [5, 4, 2, 2]])
        PM2.execute([[2, 0], [1, 3], [1, 0]])
        assert np.allclose(PM1.value, [6, 6, 5, 6])
        assert np.allclose(PM2.value, [4, 3])

class TestLinearMatrixFunction:

    def test_valid_matrix_specs(self):
        # Note: default matrix specification is None

        PM_default = ProcessingMechanism(function=LinearMatrix())
        PM_default.execute(1.0)

        assert np.allclose(PM_default.value, 1.0)

        PM_default_len_2_var = ProcessingMechanism(function=LinearMatrix(default_variable=[[0.0, 0.0]]),
                                                   default_variable=[[0.0, 0.0]])
        PM_default_len_2_var.execute([[1.0, 2.0]])

        assert np.allclose(PM_default_len_2_var.value, [[1.0, 2.0]])

        PM_default_2d_var = ProcessingMechanism(function=LinearMatrix(default_variable=[[0.0, 0.0],
                                                                                        [0.0, 0.0],
                                                                                        [0.0, 0.0]]),
                                                   default_variable=[[0.0, 0.0],
                                                                     [0.0, 0.0],
                                                                     [0.0, 0.0]])

        #  [1.0   0.0]           [1.0   0.0]
        #  [0.0   2.0]    *      [0.0   1.0]
        #  [3.0   0.0]
        PM_default_2d_var.execute([[1.0, 0.0],
                                   [0.0, 2.0],
                                   [3.0, 0.0]])

        assert np.allclose(PM_default_2d_var.value, [[1.0, 0.0],
                                                     [0.0, 2.0],
                                                     [3.0, 0.0]])

        # PM_float = ProcessingMechanism(function=LinearMatrix(matrix=4.0))
        # PM_float.execute(1.0)
        #
        # assert np.allclose(PM_float.value, 4.0)

        PM_1d_list = ProcessingMechanism(function=LinearMatrix(matrix=[4.0]))
        PM_1d_list.execute(1.0)

        assert np.allclose(PM_1d_list.value, 4.0)

        PM_2d_list = ProcessingMechanism(function=LinearMatrix(matrix=[[4.0, 5.0],
                                                                       [6.0, 7.0],
                                                                       [8.0, 9.0],
                                                                       [10.0, 11.0]],
                                                               default_variable=[[0.0, 0.0, 0.0, 0.0],
                                                                                 [0.0, 0.0, 0.0, 0.0],
                                                                                 [0.0, 0.0, 0.0, 0.0]]),
                                         default_variable=[[0.0, 0.0, 0.0, 0.0],
                                                           [0.0, 0.0, 0.0, 0.0],
                                                           [0.0, 0.0, 0.0, 0.0]])
        PM_2d_list.execute([[1.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 1.0, 0.0],
                            [0.0, 0.0, 0.0, 1.0]])

        assert np.allclose(PM_2d_list.value, [[4.0, 5.0],
                                              [8.0, 9.0],
                                              [10.0, 11.0]])

        PM_1d_array = ProcessingMechanism(function=LinearMatrix(matrix=np.array([4.0])))
        PM_1d_array.execute(1.0)

        assert np.allclose(PM_1d_array.value, 4.0)

        PM_2d_array = ProcessingMechanism(function=LinearMatrix(matrix=np.array([[4.0]])))
        PM_2d_array.execute(1.0)

        assert np.allclose(PM_2d_array.value, 4.0)

        PM_matrix = ProcessingMechanism(function=LinearMatrix(matrix=np.matrix([[4.0]])))
        PM_matrix.execute(1.0)

        assert np.allclose(PM_matrix.value, 4.0)

    def test_invalid_matrix_specs(self):

        with pytest.raises(FunctionError) as error_text:
            PM_mismatched_float = ProcessingMechanism(function=LinearMatrix(default_variable=0.0,
                                                                        matrix=[[1.0, 0.0, 0.0, 0.0],
                                                                                [0.0, 2.0, 0.0, 0.0],
                                                                                [0.0, 0.0, 3.0, 0.0],
                                                                                [0.0, 0.0, 0.0, 4.0]]),
                                                  default_variable=0.0)
        assert "Specification of matrix and/or default_variable" in str(error_text.value) and \
               "not compatible for multiplication" in str(error_text.value)

        with pytest.raises(FunctionError) as error_text:
            PM_mismatched_matrix = ProcessingMechanism(function=LinearMatrix(default_variable=[[0.0, 0.0],
                                                                                          [0.0, 0.0],
                                                                                          [0.0, 0.0]],
                                                                        matrix=[[1.0, 0.0, 0.0, 0.0],
                                                                                [0.0, 2.0, 0.0, 0.0],
                                                                                [0.0, 0.0, 3.0, 0.0],
                                                                                [0.0, 0.0, 0.0, 4.0]]),
                                                  default_variable=[[0.0, 0.0],
                                                                    [0.0, 0.0],
                                                                    [0.0, 0.0]])
        assert "Specification of matrix and/or default_variable" in str(error_text.value) and \
               "not compatible for multiplication" in str(error_text.value)

class TestProcessingMechanismStandardOutputPorts:

    def test_mean(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MEAN])
        PM1.execute([1,2,4])
        assert np.allclose(PM1.output_ports[0].value,[2.33333333])

    def test_median(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MEDIAN])
        PM1.execute([1,2,4])
        assert np.allclose(PM1.output_ports[0].value,[2])

    def test_std_dev(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[STANDARD_DEVIATION])
        PM1.execute([1,2,4])
        assert np.allclose(PM1.output_ports[0].value,[1.24721913])

    def test_variance(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[VARIANCE])
        PM1.execute([1,2,4])
        assert np.allclose(PM1.output_ports[0].value,[1.55555556])

    def test_max_val(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MAX_VAL])
        PM1.execute([1,2,-4])
        # assert np.allclose(PM1.output_ports[0].value,[0,2,0])
        assert np.allclose(PM1.output_ports[0].value,[2])

    def test_max_abs_val(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MAX_ABS_VAL])
        PM1.execute([1,2,-4])
        # assert np.allclose(PM1.output_ports[0].value,[0,0,-4])
        assert np.allclose(PM1.output_ports[0].value,[4])

    def test_max_one_hot(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MAX_ONE_HOT])
        PM1.execute([1,2,-4])
        assert np.allclose(PM1.output_ports[0].value,[0,2,0])

    def test_max_abs_one_hot(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MAX_ABS_ONE_HOT])
        PM1.execute([1,2,-4])
        assert np.allclose(PM1.output_ports[0].value,[0,0,4])

    def test_max_indicator(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MAX_INDICATOR])
        PM1.execute([1,2,-4])
        assert np.allclose(PM1.output_ports[0].value,[0,1,0])

    def test_max_abs_indicator(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[MAX_ABS_INDICATOR])
        PM1.execute([1,2,-4])
        assert np.allclose(PM1.output_ports[0].value,[0,0,1])

    def test_prob(self):
        PM1 = ProcessingMechanism(default_variable=[0,0,0], output_ports=[PROB])
        PM1.execute([1,2,4])
        assert np.allclose(PM1.output_ports[0].value,[0,0,4])



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# ****************************************  EpisodicMemoryMechanism ****************************************************

"""


Contents
--------

  * `EpisodicMemoryMechanism_Overview`
  * `EpisodicMemoryMechanism_Creation`
  * `EpisodicMemoryMechanism_Structure`
  * `EpisodicMemoryMechanism_Execution`
  * `EpisodicMemoryMechanism_Class_Reference`


.. _EpisodicMemoryMechanism_Overview:

Overview
--------

An EpisodicMemoryMechanism is an `IntegratorMechanism` that can store and retrieve content-associate pairs.
Only the content is used for determining which pairs are retrieved

.. _EpisodicMemoryMechanism_Creation:

Creating a TransferMechanism
-----------------------------

An EpisodicMemoryMechanism is created by calling its constructor with **content_size** and, optionally, **assoc_size**,
that define the shapes of the items stored in its memory.

.. _EpisodicMemoryMechanism_Structure:

Structure
---------

An EpisodicMemoryMechanism has at least one `InputPort <InputPort>`, its *CONTENT_INPUT* and,
optionally, an *ASSOC_INPUT* InputPort (if its *assoc_size* is specified and is not 0) that represent
an item to store;  a `function <EpisodicMemoryMechanism.function>` that stores and retrieves content-assoc pairs from its
memory; and at least one `OutputPort <OutputPort>`, *CONTENT_OUTPUT*, as well as a 2nd, *CONTENT_OUTPUT* if it has
an *ASSOC_INPUT* InputPort, that represent a retrieved item. The default function is a `ContentAddressableMemory` that
implements a simple form of content-addressable memory, but a custom function can be specified, so long as it meets the
following requirements:

    * It must accept a 2d array as its first argument, the first item of which is the content and the second the
      associate.

    * It must return a 2d array, the first item of which is the retrieved content and the second of which is the
      assoc with which it is associated in the `function <EpisodicMemoryMechanism.function>`\\'s `memory
      <EpisodicMemoryMechanism.memory>`.

    * It may also implement a memory attribute;  if it does, it can be accessed by the EpisodicMemoryMechanism's
      `memory <EpisodicMemoryMechanism.memory>` attribute.

.. _EpisodicMemoryMechanism_Execution:

Execution
---------

When an EpisodicMemoryMechanism is executed, its `function <EpisodicMemoryMechanism.function>` carries out
the following operations:

    * retrieves an item from its memory based on the `value <InputPort.value>` of its *CONTENT_INPUT* `InputPort`;
      if no retrieval is made, appropriately shaped zero-valued arrays are assigned to the `value
      <OutputPort.value>` of the *CONTENT_OUTPUT* and, if specified, it *ASSOC_OUTPUT* OutputPorts.
    ..
    * stores the `value <InputPort.value>` of its *CONTENT_INPUT* and, if specified, *ASSOC_INPUT* `InputPort
      <InputPort>` in its memory.
    ..
    * assigns the value of the retrieved item's content in the EpisodicMemoryMechanism's  *CONTENT_OUTPUT*
      `OutputPort`, and the value of the assoc of the retrieved item in the *ASSOC_OUTPUT* OutputPort.

    .. note::
         The order of storage and retieval is determined by the function.

         The value of the content of the item retrieved from memory (and stored in *CONTENT_OUTPUT*)
         may be different than the `value <InputPort.value>` of *CONTENT* used to retrieve the item.

         If no retrieval is made, appropriately shaped zero-valued arrays are assigned as the `value
         <OutputPort.value>` of the *CONTENT_OUTPUT* and, if specified, *ASSOC_OUTPUT* OutputPorts.

.. _EpisodicMemoryMechanism_Class_Reference:

Class Reference
---------------


"""
import warnings

import numpy as np

from psyneulink.core.components.functions.function import Function
from psyneulink.core.components.functions.statefulfunctions.memoryfunctions import ContentAddressableMemory
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.components.ports.inputport import OutputPort
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import CONTEXT, NAME, OWNER_VALUE, SIZE, VARIABLE
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.utilities import convert_to_np_array

__all__ = ['EpisodicMemoryMechanism', 'CONTENT_INPUT', 'ASSOC_INPUT', 'CONTENT_OUTPUT', 'ASSOC_OUTPUT']

CONTENT_INPUT = 'CONTENT_INPUT'
ASSOC_INPUT = 'ASSOC_INPUT'
CONTENT_OUTPUT = 'CONTENT_OUTPUT'
ASSOC_OUTPUT = 'ASSOC_OUTPUT'


class EpisodicMemoryMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


def _generate_content_input_port_spec(content_size):
    return [{NAME: CONTENT_INPUT, SIZE: content_size}]


class EpisodicMemoryMechanism(ProcessingMechanism_Base):
    """
    EpisodicMemoryMechanism(                \
        content_size=1,                     \
        assoc_size=1,                       \
        function=ContentAddressableMemory,  \
        params=None,                        \
        name=None,                          \
        prefs=None                          \
    )

    Subclass of `IntegratorMechanism <IntegratorMechanism>` that implements a `differentiable neural dictionary
    (ContentAddressableMemory)<HTML>`_.  See `Mechanism <Mechanism_Class_Reference>` for additional arguments and
    attributes.

    Arguments
    ---------

    content_size : int : default 1
        specifies length of the content stored in the `function <EpisodicMemoryMechanism.function>`\\s memory.

    assoc_size : int : default 0
        specifies length of the assoc stored in the `function <EpisodicMemoryMechanism.function>`\\s memory;
        if it is 0 (the default) then no *ASSOC_INPUT* InputPort or *ASSOC_OUTPUT* OutputPort are created.

    function : function : default ContentAddressableMemory
        specifies the function that implements a memory store and methods to store to and retrieve from it.  It
        must take as its `variable <Function.variable>` a 2d array, the first item of which is the content and the second
        the associate to be stored in its memory, and must return a 2d array that is the value of the
        content and assoc retrieved from its memory.

    Attributes
    ----------

    function : function
        function that implements storage and retrieval from a memory.

    memory : 3d array
        contains key-value pairs stored in the `function <EpisodicMemoryMechanism.function>'\\s `memory` attribute
        (if it has one).

    """

    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <EpisodicMemoryMechanism.variable>`

                    :default value: [[0]]
                    :type: ``list``

                assoc_size
                    see `assoc_size <EpisodicMemoryMechanism.assoc_size>`

                    :default value: 0
                    :type: ``int``

                content_size
                    see `content_size <EpisodicMemoryMechanism.content_size>`

                    :default value: 1
                    :type: ``int``

                function
                    see `function <EpisodicMemoryMechanism.function>`

                    :default value: `ContentAddressableMemory`
                    :type: `Function`

                input_ports
                    see `input_ports <EpisodicMemoryMechanism.input_ports>`

                    :default value: ["{name: CONTENT_INPUT, size: 1}"]
                    :type: ``list``
                    :read only: True

                output_ports
                    see `output_ports <EpisodicMemoryMechanism.output_ports>`

                    :default value: ["{name: CONTENT_OUTPUT, variable: (OWNER_VALUE, 0)}"]
                    :type: ``list``
                    :read only: True
        """
        variable = Parameter([[0]], pnl_internal=True, constructor_argument='default_variable')
        function = Parameter(ContentAddressableMemory, stateful=False, loggable=False)
        content_size = 1
        assoc_size = 0

        input_ports = Parameter(
            _generate_content_input_port_spec(content_size),
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
            parse_spec=True,
        )

        output_ports = Parameter(
            [{NAME: CONTENT_OUTPUT, VARIABLE: (OWNER_VALUE, 0)}],
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
        )

    def __init__(self,
                 content_size:int=None,
                 assoc_size:int=None,
                 function:Function=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 **kwargs):
        # Template for memory_store entries
        default_variable = [np.zeros(content_size)]
        input_ports = None
        output_ports = None

        if content_size is not None and content_size != self.defaults.content_size:
            input_ports = _generate_content_input_port_spec(content_size)

        if assoc_size is not None and assoc_size != self.defaults.assoc_size:
            try:
                input_ports.append({NAME: ASSOC_INPUT, SIZE: assoc_size})
            except AttributeError:
                input_ports = [{NAME: ASSOC_INPUT, SIZE: assoc_size}]

            output_ports = self.class_defaults.output_ports.copy()
            output_ports.append({NAME: ASSOC_OUTPUT, VARIABLE: (OWNER_VALUE, 1)})
            default_variable.append(np.zeros(assoc_size))

        super().__init__(
            default_variable=default_variable,
            function=function,
            params=params,
            name=name,
            prefs=prefs,
            input_ports=input_ports,
            output_ports=output_ports,
            **kwargs
        )

    def _execute(self, variable=None, context=None, runtime_params=None):

        value = super()._execute(variable=variable,
                                 context=context,
                                 runtime_params=runtime_params,
                                )
        # Only return content if assoc has not been specified (in which case second element of value should be empty)
        if len(value[1]) == 0:
            return np.delete(value,1)
        else:
            return value


    def _instantiate_output_ports(self, context=None):
        if len(self.input_ports) != len(self.output_ports):
            assert False, \
                f'PROGRAM ERROR: Number of {InputPort.__class__.__name__}s and ' \
                f'{OutputPort.__class__.__name__}s do not match in {self.name}'
        for i, input_port_spec, output_port_spec in zip(range(len(self.input_ports) - 1),
                                                          self.input_ports,
                                                          self.output_ports):
            if input_port_spec.value is []:
                del self.output_ports[i]

        return super()._instantiate_output_ports(context=context)

    def _parse_function_variable(self, variable, context=None):

        # If assoc has not been specified, add empty list to call to function (which expects two items in its variable)
        if len(variable) != 2:
            return convert_to_np_array([variable[0],[]])
        else:
            return variable

    @property
    def memory(self):
        """Return function's memory attribute"""
        try:
            return self.function.memory
        except:
            warnings.warning(f'Function of {self.name} (self.function.name) has no memory attribute')



from . import modulatorymechanism
from psyneulink.core.components.mechanisms.modulatory.control import controlmechanism, gating
from . import control
from . import learning

from .modulatorymechanism import *
from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import *
from .control import *
from psyneulink.core.components.mechanisms.modulatory.control.gating import *
from .learning import *

__all__ = list(control.__all__)
__all__.extend(gating.__all__)
__all__.extend(learning.__all__)
__all__.extend(modulatorymechanism.__all__)
__all__.extend(controlmechanism.__all__)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ******************************************  ControlSignal *****************************************************

"""

Contents:
---------

  * `Control_signal_Overview`
  * `ControlSignal_Creation`
      - `ControlSignal_Specification`
  * `ControlSignal_Structure`
      - `ControlSignal_Projections`
      - `ControlSignal_Modulation`
      - `ControlSignal_Allocation_and_Intensity`
      - `ControlSignal_Costs`
  * `ControlSignal_Execution`d
  * `ControlSignal_Examples`
  * `ControlSignal_Class_Reference`


.. _Control_Signal_Overview:

Overview
--------

A ControlSignal is a type of `ModulatorySignal <ModulatorySignal>` that is specialized for use with a `ControlMechanism
<ControlMechanism>` and one or more `ControlProjections <ControlProjection>`, to modify the parameter(s) of one or more
`Components <Component>`. A ControlSignal receives an `allocation <ControlSignal.allocation>` value from the
ControlMechanism to which it belongs, and uses that to compute an `intensity` (also referred to as a `control_signal`)
that is assigned as the `value <ControlProjection.ControlProjection.value>` of its ControlProjections. Each
ControlProjection conveys its value to the `ParameterPort` for the parameter it controls, which uses that value to
`modulate <ModulatorySignal_Modulation>` the `value <ParameterPort.value>` of the parameter.  A ControlSignal also
calculates a `cost`, based on its `intensity` and/or its time course, that may be used by the ControlMechanism to
adapt the ControlSignal's `allocation <ControlSignal.allocation>` in subsequent executions.

.. _ControlSignal_Creation:

Creating a ControlSignal
------------------------

A ControlSignal is created automatically whenever the parameter of a Mechanism or of its function is `specified for
control <ControlMechanism_ControlSignals>`.  ControlSignals can also be specified in the **control_signals** argument
of the constructor for a `ControlMechanism <ControlMechanism>`, as well as in the `specification of the parameter
<ParameterPort_Specification>` that the ControlSignal is intended to modulate (also see `Modualory Specification
<ParameterPort_Modulatory_Specification>`.  Although a ControlSignal can also be  created on its own using its
constructor (or any of the other ways for `creating an OutputPort <OutputPorts_Creation>`), this is usually not
necessary nor is it advisable, as a ControlSignal has dedicated components and requirements for configuration that
must be met for it to function properly.

.. _ControlSignal_Specification:

*Specifying ControlSignals*
~~~~~~~~~~~~~~~~~~~~~~~~~~~

When a ControlSignal is specified in the **control_signals** argument of the constructor for a `ControlMechanism
<ControlMechanism>`, the parameter(s) to be controlled must be specified.  If other attributes of the ControlSignal
need to be specified (e.g., one or more of its `cost functions <ControlSignal_Costs>`), then the Constructor for the
ControlSignal can be used or a `port specification dictionary <Port_Specification>`, in which the parameter(s) to be
controlled in the **projections** argument or *PROJECTIONS* entry, respectively, using any of the forms below.
For convenience, the parameters can also be specified on their own in the **control_signals** argument of the
ControlMechanism's constructor, in which case a default ControlSignal will be created for each.  In all cases, any
of the following can be use to specify the parameter(s) to be controlled:

  * **ParameterPort** (or list of them) -- for the Mechanism(s) to which the parameter(s) belong;

  * **2-item tuple:** *(parameter name or list of them>, <Mechanism>)* -- the 1st item must be the name of the
    parameter (or list of parameter names), and the 2nd item the Mechanism to which it (they) belong(s); this is a
    convenience format, that is simpler to use than a specification dictionary (see above), but precludes
    specification of any `parameters <ControlSignal_Structure>` for the ControlSignal.

  .. _ControlSignal_Specification_Dictionary:

  * **specification dictionary** -- this is an abbreviated form of `port specification dictionary
    <Port_Specification>`, in which the parameter(s) to be controlled can be specified in either of the two
    following ways:

    * for controlling a single parameter, the dictionary can have the following two entries:

        * *NAME*: str
            the string must be the name of the parameter to be controlled;

        * *MECHANISM*: Mechanism
            the Mechanism must be the one to the which the parameter to be controlled belongs.

    * for controlling multiple parameters, the dictionary can have the following entry:

        * <str>:list
            the string used as the key specifies the name to be used for the ControlSignal,
            and each item of the list must be a `specification of a parameter <ParameterPort_Specification>` to be
            controlled by the ControlSignal (and that will receive a `ControlProjection` from it).
  ..

.. _ControlSignal_Structure:

Structure
---------

A ControlSignal is owned by an `ControlMechanism <ControlMechanism>`, and controls the parameters of one or more
Components by modulating the `function <ParameterPort.function>` of the `ParameterPort` that determines the value
of each of the parameters that it control.  Its operation is governed by several attributes of the ControlSignal,
that are described below.

.. _ControlSignal_Projections:

*Projections*
~~~~~~~~~~~~~

When a ControlSignal is created, it can be assigned one or more `ControlProjections <ControlProjection>`, using either
the **projections** argument of its constructor, or in an entry of a dictionary assigned to the **params** argument
with the key *PROJECTIONS*.  These will be assigned to its `efferents  <ModulatorySignal.efferents>` attribute.  See
`Port Projections <Port_Projections>` for additional details concerning the specification of Projections when
creating a Port.

.. note::
   Although a ControlSignal can be assigned more than one `ControlProjection`, all of those Projections will receive
   the same `value <ControlProjection.value>` (based on the `intensity` of that ControlSignal), and use the same
   form of `modulation <ControlSignal_Modulation>`.  Thus, for them to be meaningful, they should project to
   ParameterPorts for parameters that are meaningfully related to one another (for example, the threshold parameter
   of multiple `DDM` Mechanisms).

.. _ControlSignal_Modulation:

*Modulation*
~~~~~~~~~~~~

A ControlSignal has a `modulation <ModulatorySignal.modulation>` attribute that determines how its ControlSignal's
`value <ControlSignal.value>` is used by the Ports to which it projects to modify their `value <Port_Base.value>` \\s
(see `ModulatorySignal_Modulation` for an explanation of how the `modulation <ModulatorySignal.modulation>` attribute is
specified and used to modulate the `value <Port_Base.value>` of a Port). The `modulation <ModulatorySignal.modulation>`
attribute can be specified in the **modulation** argument of the constructor for a ControlSignal, or in a specification
dictionary as described `above <ControlSignal_Specification>` (see `ModulatorySignal_Types` for forms of specification).
If it is not specified, its default is the value of the `modulation <ControlMechanism.modulation>` attribute of the
ControlMechanism to which the ControlSignal belongs (which is the same for all of the ControlSignals belonging to
that ControlMechanism).  The value of the `modulation <ModulatorySignal.modulation>` attribute of a ControlSignal is
used by all of the `ControlProjections <ControlProjection>` that project from that ControlSignal.

.. _ControlSignal_Allocation_and_Intensity:

*Allocation, Function and Intensity*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

*Allocation (variable)*. A ControlSignal is assigned an `allocation <ControlSignal>` by the ControlMechanism to which
it belongs. Some ControlMechanisms sample different allocation values for their ControlSignals to determine which to
use (e.g., `OptimizationControlMechanism`);  in those cases, they use each ControlSignal's
`allocation_samples <ControlSignal.allocation_samples>` attribute (specified in the **allocation_samples** argument
of the ControlSignal's constructor) to determine the allocation values to sample for that ControlSignal.  A
ControlSignal's `allocation <ControlSignal>` attribute contains the value assigned to it by the ControlMechanism
at the end of the previous `TRIAL <TimeScale.TRIAL>` (i.e., when the ControlMechanism last executed --  see
`ControlMechanism Execution <ControlMechanism_Execution>`); its value from the previous `TRIAL <TimeScale.TRIAL>` is
assigned to the `last_intensity` attribute.

*Function*. A ControlSignal's `allocation <ControlSignal.allocation>` serves as its `variable
<ModulatorySignal.variable>`, and is used by its `function <ControlSignal.function>` to generate an `intensity`.
The default `function <ControlSignal.function>` for a ControlSignal is `TransferWithCosts`.  This is a
`Function` that supplements its core `TransferFunction` (specified by its `transfer_fct
<TransferWithCosts.transfer_fct>` with a set of cost functions that can be used to compute the ControlSignal's `cost
attributes <ControlSignal_Costs>`.  The default `transfer_fct <TransferWithCosts.transfer_fct>`> for TransferWithCosts
is an identity function (`Linear` with `slope <Linear.slope>` \\=1 and `intercept <Linear.intercept>`\\=0), that simply
assigns the ControlSignal's `allocation <ControlSignal.allocation>` as its `intensity <ControlSignal.intensity>`.
However, the TransferWithCosts function can be specified as **function** argument in the ControlSignal's constructor,
with a different function specified for the **transfer_fct** argument of the TransferWithCosts's constructor (e.g.,
`Exponential`, or any other function that takes and returns a scalar value or 1d array).  The TransferWithCosts' `cost
functions <TransferWithCosts_Cost_Functions>` can also be assigned using its own constructor.  A function other than
TransferWithCosts can also be assigned as the ControlSignal's `function <ControlSignal.function>`, however in that
case, the ControlSignal's costs can't be computed and will all be assigned None.

*Intensity (value)*. The result of the function is assigned as the value of the ControlSignal's `intensity`
attribute, which serves as the ControlSignal's `value <ControlSignal.value>` (also referred to as `control_signal`).
The `intensity` is used by its `ControlProjection(s) <ControlProjection>` to modulate the parameter(s) for which the
ControlSignal is responsible. The ControlSignal's `intensity` attribute  reflects its value for the current `TRIAL
<TimeScale.TRIAL>`; its value from the previous `TRIAL <TimeScale.TRIAL>` is assigned to the `last_intensity` attribute.

.. _ControlSignal_Costs:

*Costs and Cost Functions*
~~~~~~~~~~~~~~~~~~~~~~~~~~

A ControlSignal has a `cost <ControlSignal.cost>` attribute that may be used by the ControlMechanism to which it
belongs to determine its `allocation <ControlSignal.allocation>`.  The value of the `cost <ControlSignal.cost>`
is computed from the ControlSignal's `intensity` using one or more of four cost functions.  These are only
available if the ControlSignal's `function <ControlSignal.function>` is `TransferWithCosts` (which it is by default),
and are actually functions that belong to the `TransferWithCosts` Function (see `TransferWithCosts_Cost_Functions`).
Three of these compute different types of cost, and a fourth combines them, the result of which is assigned as the
ControlSignal's `cost <ControlSignal.cost>` attribute.  The ControlSignal has attributes that reference each of the
TransferWithCosts' cost functions and their attributesas listed below:

    * `intensity_cost` - calculated by the `intensity_cost_fct <TransferWithCosts.intensity_cost_fct>` based on the
      current `intensity` of the ControlSignal, and can be referenced by the ControlSignal's `intensity_cost_function
      <ControlSignal.intensity_cost_function>` attribute.
    ..
    * `adjustment_cost` - calculated by the `adjustment_cost_fct <TransferWithCosts.adjustment_cost_fct>` based on
      a change in the ControlSignal's `intensity` from its last value, and can be referenced by the ControlSignal's
      `adjustment_cost_function <ControlSignal.adjustment_cost_function>` attribute.
    ..
    * `duration_cost` - calculated by the `duration_cost_fct <TransferWithCosts.duration_cost_fct>` based on an
      integral of the ControlSignal's `cost <ControlSignal.cost>`, and can be referenced by the ControlSignal's
      `duration_cost_function <ControlSignal.duration_cost_function>` attribute.
    ..
    * `cost` - calculated by the `combine_costs_fct <TransferWithCosts.combine_costs_fct>` that combines the results
      of any cost functions that are enabled, and can be referenced by the ControlSignal's `duration_cost_function
      <ControlSignal.combine_costs_function>` attribute.

Which of the cost functions are used can be specified in the **costs_options** argument ControlSignal's constructor,
which is passed to the corresponding argument of the TransferWithCosts constructor (where it can also be specified).
The costs functions can also be enabled and disabled using the TransferWithCosts function's `enable_costs
<TransferWithCosts.enable_costs>`,  `disable_costs <TransferWithCosts.disable_costs>`, `toggle_cost
<TransferWithCosts.toggle_cost>` and `assign_costs <TransferWithCosts.assign_costs>` methods.  All of these
take one or more values of `CostFunctions`.  The `combine_costs_function <ControlSignal.combine_costs_function>` is used
to combine the values generated by the enabled cost functions, and the result is assigned as the ControlSignal's `cost
<ControlSignal.cost>` attribute.  By default, the `combine_costs_function <ControlSignal.combine_costs_function>` sums
the results of the enabled cost functions. However, this can be modified by specifying the `combine_costs_function
<ControlSignal.combine_costs_function>` by specifying the ControlSignal's  `function <ControlSignal.function>` using
the constructor for the TransferWithCosts function. The parameters of the ControlSignal's cost functions can also be
modulated by another `ControlMechanism` (see `example <ControlSignal_Example_Modulate_Costs>` below).

    COMMENT:
    .. _ControlSignal_Toggle_Costs:

    *Enabling and Disabling Cost Functions*.  Any of the cost functions (except the `combine_costs_function`) can
    be enabled or disabled using the `toggle_cost` method to turn it `ON` or `OFF`. If it is disabled, that
    component of the cost is not included in the ControlSignal's `cost` attribute.  A cost function  can  also be
    permanently disabled for the ControlSignal by assigning it's attribute `None`.  If a cost function is permanently
    disabled for a ControlSignal, it cannot be re-enabled using `toggle_cost`.
    COMMENT

.. _ControlSignal_Execution:

Execution
---------

A ControlSignal cannot be executed directly.  It is executed whenever the `ControlMechanism <ControlMechanism>` to
which it belongs is executed.  When this occurs, the ControlMechanism provides the ControlSignal with an `allocation
<ControlSignal.allocation>`, that is used by its `function <ControlSignal.function>` to compute its `intensity` for that
`TRIAL <TimeScale.TRIAL>`.  The `intensity` is used by the ControlSignal's `ControlProjections <ControlProjection>` to
set the `value <ParameterPort.value>`\\(s) of the `ParameterPort(s) <ParameterPort>` to which the ControlSignalprojects.

Recall that the ParameterPort value is referenced anywhere that the controlled parameter is used in computation, and
that it does not update until the component to which the ParameterPort belongs executes. If the distinction between the
base value stored in the parameter attribute (i.e. MyTransferMech.function.gain) and the value of the
ParameterPort is unfamiliar, see `Parameter Port documentation <ParameterPort>` for more details, or see
`ModulatorySignal_Modulation` for a detailed description of how modulation operates.

The ControlSignal's `intensity` is also used  by its `cost functions <ControlSignal_Costs>` to compute its `cost`
attribute. That is used by some ControlMechanisms, along with the ControlSignal's `allocation_samples` attribute, to
evaluate a `control_allocation <ControlMechanism.control_allocation>`, and adjust the ControlSignal's `allocation
<ControlSignal.allocation>` for the next `TRIAL <TimeScale.TRIAL>`.

.. note::
   The changes in a parameter in response to the execution of a ControlMechanism are not applied until the Mechanism
   with the parameter being controlled is next executed; see `Lazy Evaluation <Component_Lazy_Updating>` for an
   explanation of "lazy" updating).

.. _ControlSignal_Examples:

Examples
--------

As explained `above <ControlSignal_Creation>` (and in `ControlMechanism_ControlSignals`), a ControlSignal can be
specified either where the parameter it controls is specified, or in the constructor of the ControlMechanism to
which the ControlSignal belongs (i.e., that is responsible for the control).  Which of these to use is largely an
aesthetic matter -- that is, where you wish the specification of control to appear. Examples of each are provided below.

.. note::
  If a ControlSignal is specified where the parameter it controls is specified, the ControlSignal is
  not implemented until the `Component` to which the parameter belongs is assigned to a `Composition`
  (see `ControlMechanism_ControlSignals`).

.. _ControlSignal_Example_Specify_with_Parameter:

*Specify a ControlSignal where a parameter is specified*

The simplest way to do this is to specify the ControlSignal by class, or using the keyword *CONTROL*, in a
`tuple specification <ParameterPort_Tuple_Specification>` for the parameter, as in the following examples:

    >>> from psyneulink import *
    >>> my_mech = ProcessingMechanism(function=Logistic(bias=(1.0, ControlMechanism))) #doctest: +SKIP
    or
    >>> my_mech = ProcessingMechanism(function=Logistic(bias=(1.0, CONTROL)))

Both of these assign a ControlSignal to the `bias <Logistic.gain>` parameter of the `Logistic` Function used by a
`ProcessingMechanism`. This creates a default ControlSignal, with a `ControlProjection` that projects to the
ProcessingMechanism's `ParameterPort` for the `bias <Logistic.bias>` parameter of its `Logistic` Function.  The
default value of a ControlSignal's `modulation <ModulatorySignal.modulation>` attribute is *MULTIPLICATIVE*,
so that it will multiply the value of the `bias <Logistic.bias>` parameter. When the ProcessingMechanism executes,
the Logistic Function will use the value of the ControlSignal as its bias parameter.

.. _ControlSignal_Example_Modulation:

If attributes of the ControlSignal must be specified, then its constructor can be used.  For example, ordinarily a
ControlSignal modulates the `MULTIPLICATIVE_PARAM <ModulatorySignal_Types>` of a `Port's <Port>` `function
<Port_Base.function>`.  However, this can be changed by using the ControlSignal's constructor to specify its
**modulation** argument, as follows::

    >>> my_mech = ProcessingMechanism(function=Logistic(gain=(1.0, ControlSignal(modulation=ADDITIVE))))

This specifies that the value of the ControlSignal should be added to, rather than multiply the value of the `gain
<Logistic.gain>` parameter of the Logistic function.  Note that the **modulation** argument determines how to modify a
*parameter* of the *Logistic* Function (in this case, its `gain <Logistic.gain>` parameter), and its input directly;
that is, in this example, the value of the ControlSignal is added to the *gain parameter* of the Logistic function,
*not* to its `variable <Lo˚gistic.variable>`.  If the value of the ControlSignal's **modulation** argument had been
*OVERRIDE*, then the ControlSignal's value would have been used as (i.e., it would have replaced) the ˚value of the
*Logistic* Function's `gain <Logistic.gain>` parameter, rather than being added to it.

*Specify ControlSignals in a ControlMechanism's constructor*

Parameters can also be specified for control in constructor of the `ControlMechanism <ControlMechanism>` that controls
them.  This is done in the **control_signals** argument of the ControlMechanism's constructor, by specifying a
ControlSignal for each parameter to be controlled.  The following example shows several ways in which the ControlSignal
can be specified (see `ControlSignal_Specification` for additional details)::

    >>> mech_A = ProcessingMechanism()
    >>> mech_B = ProcessingMechanism(function=Logistic)
    >>> ctl_mech = ControlMechanism(control_signals=[(INTERCEPT, mech_A),
    ...                                              ControlSignal(modulates=(GAIN, mech_B),
    ...                                                            cost_options=CostFunctions.INTENSITY),
    ...                                              {
    ...                                                  NAME: BIAS,
    ...                                                  MECHANISM: mech_B,
    ...                                                  MODULATION: ADDITIVE
    ...                                              }])

Here, ``ctl_mech`` modulates the `intercept <Linear.intercept>` parameter of ``mech_A``'s `Linear` Function
(the default `function <Mechanism_Base.function>` for a `ProcessingMechanism`), and the `bias <Logistic.bias>`
and `gain <Logistic.gain>` parameters of ``mech_B``'s `Logistic` Function.  The first ControlSignal is specified
using a 2-item tuple (the simplest way to do so); the second uses the ControlSignal's constructor (allowing another
parameter to be specified -- here, its `cost_options <ControlSignal_Costs>`); and the third uses a specification
dictionary (which supports additional options; see `above <ControlSignal_Specification_Dictionary>`).

.. _ControlSignal_Example_Modulate_Costs:

*Modulate the parameters of a ControlSignal's cost function*.

ControlSignals can be used to modulate the parameters of other ControlSignals.  This is done using the **modulation**
argument, but instead of using the keyword for a `generic form of modulation <ModulatorySignal_Types>` (as in the
example `above <ControlSignal_Example_Modulation>`), a parameter of the ControlSignal's `function
<ControlSignal.function>` is used.  For example, the following shows how the parameters of a ControlSignal's `cost
function <ControlSignal_Costs>` can be modulated by another ControlSignal::

  >>> from psyneulink import *
  >>> mech = ProcessingMechanism(name='my_mech')
  >>> ctl_mech_A = ControlMechanism(monitor_for_control=mech,
  ...                               control_signals=ControlSignal(modulates=(INTERCEPT,mech),
  ...                                                             cost_options=CostFunctions.INTENSITY))
  >>> ctl_mech_B = ControlMechanism(monitor_for_control=mech,
  ...                               control_signals=ControlSignal(modulates=ctl_mech_A.control_signals[0],
  ...                                                             modulation=INTENSITY_COST_FCT_MULTIPLICATIVE_PARAM))
  >>> comp = Composition()
  >>> pway = comp.add_linear_processing_pathway(pathway=[mech,
  ...                                              ctl_mech_A,
  ...                                              ctl_mech_B
  ...                                              ])
  >>> pway.pathway
  [(ProcessingMechanism my_mech)]

Here, the `ControlSignal` of ``ctl_mech_A`` is configured to monitor the output of ``mech``, modulate the
the `intercept <Linear.intercept>` parameter of its `function <Mechanism_Base.function>` (which is a `Linear` by
default), and to implement its `intensity_cost_fct <TransferWithCosts.intensity_cost_fct>` (using the
**cost_options** argument of the ControlSignal's constructor).  ``ctl_mech_B`` is configured to also monitor
``mech``, but to modulate the `multiplicative_param <Function_Modulatory_Params>` of the `intensity_cost_fct
<TransferWithCosts.intensity_cost_fct>` of ``ctl_mech_A``\\s ControlSignal.  The default for the `intensity_cost_fct
<TransferWithCosts.intensity_cost_fct>` is `Exponential`, the `multiplicative_param <Function_Modulatory_Params>`
of which is `rate <Exponential>`.  (Note that the pathway returned from the call to `add_linear_processing_pathway
<Composition.add_linear_processing_pathway>` contains only ``my_mech``, since that is the only `ProcessingMechanism`
in the `Pathway`. When the ``comp`` is run with an input of ``3``, since the default `function
<Mechanism_Base.function>` for ``mech`` is `Linear` with a `slope <Linear.slope>` of 1 and an `intercept <Linear>`
of 0, its output is also ``3``, which is used by both ``ctl_mech_A`` and ``ctl_mech_B`` as their `allocation
<ControlSignal.allocation>`.  Since the ControlSignals of both use their default `function <ControlSignal>` ——
TransferWithCosts with an identity function as its `transfer_fct <TransferWithCosts.transfer_fct>` -- the
`intensity <ControlMechanism.intensity>` of both is the same as their ``allocation`` (i.e., ``3``).  This is used
as the input to the `Exponential` function used as `intensity_cost_function <ControlSignal.intensity_cost_function>`
of ``ctl_mech_A``.  However, since the `rate <Exponential.rate>` of that function is modulated by ``ctl_mech_B``,
the `intensity <ControlSignal.intensity>` of which is also ``3``, the value of the `intensity_cost
<ControlSignal.intensity_cost>` for ``ctl_mech_A`` is e ^ (allocation (3) * value of ctl_mech_B (also 3)) = e^9,
as shown below::

    >>> comp.run(inputs={mech:[3]}, num_trials=2)
    [array([3.])]
    >>> ctl_mech_A.control_signals[0].intensity_cost
    array([8103.08392758])

The Composition must be executed for 2 trials to see this, since the `value <ControlProjection.value>` computed
by ``ctl_mech_B`` must be computed on the first trial before it can have its effect on ``ctl_mech_A`` on the next
(i.e., second) trial (see noted under `ControlSignal_Execution`).

.. _ControlSignal_Class_Reference:

Class Reference
---------------

"""

import numpy as np
import typecheck as tc

# FIX: EVCControlMechanism IS IMPORTED HERE TO DEAL WITH COST FUNCTIONS THAT ARE DEFINED IN EVCControlMechanism
#            SHOULD THEY BE LIMITED TO EVC??
from psyneulink.core.components.functions.combinationfunctions import Reduce
from psyneulink.core.components.functions.function import is_function_type
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import SimpleIntegrator
from psyneulink.core.components.functions.transferfunctions import Exponential, Linear, CostFunctions, TransferWithCosts
from psyneulink.core.components.ports.modulatorysignals.modulatorysignal import ModulatorySignal
from psyneulink.core.components.ports.outputport import SEQUENTIAL, _output_port_variable_getter
from psyneulink.core.components.ports.port import Port_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.defaults import defaultControlAllocation
from psyneulink.core.globals.keywords import \
    ALLOCATION_SAMPLES, CONTROLLED_PARAMS, CONTROL_PROJECTION, CONTROL_SIGNAL, \
    INPUT_PORT, INPUT_PORTS, \
    OUTPUT_PORT, OUTPUT_PORTS, OUTPUT_PORT_PARAMS, \
    PARAMETER_PORT, PARAMETER_PORTS, \
    PROJECTION_TYPE, RECEIVER, SUM, FUNCTION
from psyneulink.core.globals.parameters import Parameter, get_validator_by_function
from psyneulink.core.globals.sampleiterator import is_sample_spec
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel
from psyneulink.core.globals.utilities import \
    is_numeric, iscompatible, kwCompatibilityLength, kwCompatibilityNumeric, kwCompatibilityType, convert_all_elements_to_np_array
from psyneulink.core.globals.sampleiterator import SampleSpec, SampleIterator

__all__ = ['ControlSignal', 'ControlSignalError', 'COST_OPTIONS']

from psyneulink.core.components.functions.transferfunctions import \
    ENABLED_COST_FUNCTIONS, \
    INTENSITY_COST, INTENSITY_COST_FUNCTION, ADJUSTMENT_COST, ADJUSTMENT_COST_FUNCTION, \
    DURATION_COST, DURATION_COST_FUNCTION, COMBINED_COSTS, COMBINE_COSTS_FUNCTION, costFunctionNames

COST_OPTIONS = 'cost_options'

def _cost_options_getter(owning_component=None, context=None):
    try:
        return getattr(owning_component.function.parameters, ENABLED_COST_FUNCTIONS)._get(context)
    except (TypeError, IndexError):
        return None


def _cost_options_setter(value, owning_component=None, context=None):
    if hasattr(owning_component, "function") and owning_component.function:
        if hasattr(owning_component.function.parameters, ENABLED_COST_FUNCTIONS):
            getattr(owning_component.function.parameters, ENABLED_COST_FUNCTIONS)._set(value, context)
    return value


def _intensity_cost_getter(owning_component=None, context=None):
    try:
        return getattr(owning_component.function.parameters, INTENSITY_COST)._get(context)
    except (TypeError, IndexError):
        return None


def _adjustment_cost_getter(owning_component=None, context=None):
    try:
        return getattr(owning_component.function.parameters, ADJUSTMENT_COST)._get(context)
    except (TypeError, IndexError):
        return None


def _duration_cost_getter(owning_component=None, context=None):
    try:
        return getattr(owning_component.function.parameters, DURATION_COST)._get(context)
    except (TypeError, IndexError):
        return None


def _cost_getter(owning_component=None, context=None):
    try:
        return getattr(owning_component.function.parameters, COMBINED_COSTS)._get(context)
    except (TypeError, IndexError):
        return None


def _intensity_cost_function_getter(owning_component=None, context=None):
    try:
        return owning_component.function.parameters.intensity_cost_fct._get(context)
    except (TypeError, IndexError):
        return None


def _adjustment_cost_function_getter(owning_component=None, context=None):
    try:
        return owning_component.function.parameters.adjustment_cost_fct._get(context)
    except (TypeError, IndexError):
        return None


def _duration_cost_function_getter(owning_component=None, context=None):
    try:
        return owning_component.function.parameters.duration_cost_fct._get(context)
    except (TypeError, IndexError):
        return None


def _combine_costs_function_getter(owning_component=None, context=None):
    try:
        return owning_component.function.parameters.combine_costs_fct._get(context)
    except (TypeError, IndexError):
        return None


class ControlSignalError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value


    def __str__(self):
        return repr(self.error_value)


class ControlSignal(ModulatorySignal):
    """
    ControlSignal(                                                 \
        default_allocation=defaultControlAllocation,               \
        function=TransferWithCosts,                                \
        costs_options=None,                                        \
        intensity_cost_function=Exponential,                       \
        adjustment_cost_function=Linear,                           \
        duration_cost_function=IntegratorFunction,                 \
        combine_costs_function=Reduce(operation=SUM),              \
        allocation_samples=self.class_defaults.allocation_samples, \
        modulates=None,                                            \
        projections=None)

    A subclass of `ModulatorySignal <ModulatorySignal>` used by a `ControlMechanism <ControlMechanism>` to
    modulate the parameter(s) of one or more other `Mechanisms <Mechanism>`.  See `ModulatorySignal
    <ModulatorySignal_Class_Reference>` for additional arguments and attributes.

    Arguments
    ---------

    default_allocation : scalar, list or np.ndarray : defaultControlAllocation
        specifies the template and default value used for `allocation <ControlSignal.allocation>`;  must match the
        shape of each item specified in `allocation_samples <ControlSignal.allocation_samples>`.

    function : Function or method : default TransferWithCosts(transfer_fct=Linear(slope=1, intercept=0))
        specifies the function used to determine the `intensity` of the ControlSignal from its `allocation`;
        must be TransferWithCosts or a subclass of that, or one that meets the requirements described see `above
        <ControlSignal_Function>`);  see `function <ControlSignal.function>` for default behavior.

    cost_options : CostFunctions or List[CostsFunctions] : None
        specifies the cost components to include in the computation of the ControlSignal's `cost <ControlSignal.cost>`.

    intensity_cost_function : Optional[TransferFunction] : default Exponential
        specifies the function used to calculate the contribution of the ControlSignal's `intensity` to its
        `cost <ControlSignal.cost>`.

    adjustment_cost_function : Optional[TransferFunction] : default Linear
        specifies the function used to calculate the contribution of the change in the ControlSignal's `intensity`
        (from its `last_intensity` value) to its `cost <ControlSignal.cost>`.

    duration_cost_function : IntegratorFunction : default IntegratorFunction
        specifies the function used to calculate the contribution of the ControlSignal's duration to its
        `cost <ControlSignal.cost>`.

    combine_costs_function : function : default `Reduce(operation=SUM) <Function.Reduce>`
        specifies the function used to combine the results of any cost functions that are enabled, the result of
        which is assigned as the ControlSignal's `cost <ControlSignal.cost>` attribute.

    allocation_samples : list, 1d array, or SampleSpec : default SampleSpec(0.1, 1, 0.1)
        specifies the values used by the ControlSignal's `owner <ModulatorySignal.owner>` to determine its
        `control_allocation <ControlMechanism.control_allocation>` (see `ControlSignal_Execution`).

    modulates : list of Projection specifications
        specifies the `ControlProjection(s) <ControlProjection>` to be assigned to the ControlSignal, and that will be
        listed in its `efferents <ModulatorySignal.efferents>` attribute (see `ControlSignal_Projections` for additional
        details).

    Attributes
    ----------

    allocation : float : default: defaultControlAllocation
        value assigned by the ControlSignal's `owner <ModulatorySignal.owner>`, and used as the `variable
        <Projection_Base.variable>` of its `function <ControlSignal.function>` to determine the ControlSignal's
        `ControlSignal.intensity`.
    COMMENT:
    FOR DEVELOPERS:  Implemented as an alias of the ControlSignal's variable Parameter
    COMMENT

    last_allocation : float
        value of `allocation` in the previous execution of ControlSignal's `owner <ModulatorySignal.owner>`.

    allocation_samples : SampleIterator
        `SampleIterator` created from **allocation_samples** specification and used to generate a set of values to
        sample by the ControlSignal's `owner <ModulatorySignal.owner>` when determining its `control_allocation
        <ControlMechanism.control_allocation>`.

    function : TransferWithCosts
        converts `allocation` into the ControlSignal's `intensity`.  The default is the identity function, which
        assigns the ControlSignal's `allocation` as its `intensity`, and does not calculate any costs.  See
        `ControlSignals_Function` for additional details.

    value : float
        result of the ControlSignal's `function <ControlSignal.function>`; same as `intensity` and `control_signal`.

    intensity : float
        result of the ControlSignal's `function <ControlSignal.function>`;
        assigned as the value of the ControlSignal's ControlProjection, and used to modify the value of the parameter
        to which the ControlSignal is assigned; same as `control_signal <ControlSignal.control_signal>`.

    last_intensity : float
        the `intensity` of the ControlSignal on the previous execution of its `owner <ModulatorySignal.owner>`.

    control_signal : float
        result of the ControlSignal's `function <ControlSignal.function>`; same as `intensity`.

    cost_options : CostFunctions or None
        boolean combination of currently assigned `CostFunctions`. Specified initially in **costs** argument of
        ControlSignal's constructor;  can be modified using the `assign_cost_options` method.

    intensity_cost_function : TransferFunction : default default Exponential
        calculates `intensity_cost` from the current value of `intensity`. It can be any `TransferFunction`, or any
        other function that takes and returns a scalar value. The default is `Exponential`.  It can be disabled
        permanently for the ControlSignal by assigning `None`.

    intensity_cost : float
        cost associated with the current `intensity`.

    adjustment_cost_function : TransferFunction : default Linear
        calculates `adjustment_cost` based on the change in `intensity` from  `last_intensity`.  It can be any
        `TransferFunction`, or any other function that takes and returns a scalar value. It can be disabled
        permanently for the ControlSignal by assigning `None`.

    adjustment_cost : float
        cost associated with last change to `intensity`.

        .. note::

        A ControlSignal's `adjustment_cost`, and its `adjustment_cost_function` are distinct from the
        `reconfiguration_cost <ControlMechanism.reconfiguration_cost>` and `compute_reconfiguration_cost
        <ControlMechanism.compute_reconfiguration_cost` function of the `ControlMechanism` to which the
        ControlSignal belongs (see `ControlMechanism Reconfiguration Cost <ControlMechanism_Reconfiguration_Cost>`
        for additional details).

    duration_cost_function : IntegratorFunction : default Linear
        calculates an integral of the ControlSignal's `cost`.  It can be any `IntegratorFunction`, or any other
        function that takes a list or array of two values and returns a scalar value. It can be disabled permanently
        for the ControlSignal by assigning `None`.

    duration_cost : float
        intregral of `cost`.

    combine_costs_function : function : default Reduce(operation=SUM)
        combines the results of all cost functions that are enabled, and assigns the result to `cost`.
        It can be any function that takes an array and returns a scalar value.

    cost : float
        combined result of all `cost functions <ControlSignal_Costs>` that are enabled.

    efferents : [List[ControlProjection]]
        a list of the `ControlProjections <ControlProjection>` assigned to (i.e., that project from) the ControlSignal.

    """

    #region CLASS ATTRIBUTES

    componentType = CONTROL_SIGNAL
    paramsType = OUTPUT_PORT_PARAMS

    class Parameters(ModulatorySignal.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <ControlSignal.variable>`

                    :default value: numpy.array([1.])
                    :type: ``numpy.ndarray``

                value
                    see `value <ControlSignal.value>`

                    :default value: numpy.array([1.])
                    :type: ``numpy.ndarray``
                    :read only: True

                adjustment_cost
                    see `adjustment_cost <ControlSignal.adjustment_cost>`

                    :default value: 0
                    :type: ``int``
                    :read only: True

                adjustment_cost_function
                    see `adjustment_cost_function <ControlSignal.adjustment_cost_function>`

                    :default value: `Linear`
                    :type: `Function`

                allocation_samples
                    see `allocation_samples <ControlSignal.allocation_samples>`

                    :default value: None
                    :type:

                combine_costs_function
                    see `combine_costs_function <ControlSignal.combine_costs_function>`

                    :default value: `Reduce`
                    :type: `Function`

                cost
                    see `cost <ControlSignal.cost>`

                    :default value: None
                    :type:
                    :read only: True

                cost_options
                    see `cost_options <ControlSignal.cost_options>`

                    :default value: CostFunctions.INTENSITY
                    :type: `CostFunctions`

                duration_cost
                    see `duration_cost <ControlSignal.duration_cost>`

                    :default value: 0
                    :type: ``int``
                    :read only: True

                duration_cost_function
                    see `duration_cost_function <ControlSignal.duration_cost_function>`

                    :default value: `SimpleIntegrator`
                    :type: `Function`

                function
                    see `function <ControlSignal.function>`

                    :default value: `TransferWithCosts`
                    :type: `Function`

                intensity_cost
                    see `intensity_cost <ControlSignal.intensity_cost>`

                    :default value: None
                    :type:
                    :read only: True

                intensity_cost_function
                    see `intensity_cost_function <ControlSignal.intensity_cost_function>`

                    :default value: `Exponential`
                    :type: `Function`

                transfer_function
                    see `transfer_function <ControlSignal.transfer_function>`

                    :default value: `Linear`
                    :type: `Function`
        """
        # FIX: if the specification of this getter is happening in several other classes, should consider
        #      refactoring Parameter to allow individual attributes to be inherited, othwerise, leaving this is an
        #      isolated case
        variable = Parameter(
            np.array([defaultControlAllocation]),
            aliases='allocation',
            getter=_output_port_variable_getter,
            pnl_internal=True, constructor_argument='default_variable'
        )

        function = Parameter(TransferWithCosts, stateful=False, loggable=False)
        transfer_function = Parameter(
            Linear,
            stateful=False,
            loggable=False,
            function_parameter=True
        )

        value = Parameter(
            np.array([defaultControlAllocation]),
            read_only=True,
            aliases=['intensity'],
            pnl_internal=True,
            history_min_length=1
        )
        allocation_samples = Parameter(None, modulable=True)

        cost_options = Parameter(
            CostFunctions.DEFAULTS,
            getter=_cost_options_getter,
            setter=_cost_options_setter,
            valid_types=(CostFunctions, list)
        )
        intensity_cost = Parameter(
            None,
            read_only=True,
            getter=_intensity_cost_getter,
            function_parameter=True
        )
        adjustment_cost = Parameter(
            0,
            read_only=True,
            getter=_adjustment_cost_getter,
            function_parameter=True
        )
        duration_cost = Parameter(
            0,
            read_only=True,
            getter=_duration_cost_getter,
            function_parameter=True
        )

        cost = Parameter(None, read_only=True, getter=_cost_getter)

        intensity_cost_function = Parameter(
            Exponential,
            stateful=False,
            loggable=False,
            function_parameter=True,
            getter=_intensity_cost_function_getter
        )
        adjustment_cost_function = Parameter(
            Linear,
            stateful=False,
            loggable=False,
            function_parameter=True,
            getter=_adjustment_cost_function_getter
        )
        duration_cost_function = Parameter(
            SimpleIntegrator,
            stateful=False,
            loggable=False,
            function_parameter=True,
            getter=_duration_cost_function_getter
        )
        combine_costs_function = Parameter(
            Reduce,
            stateful=False,
            loggable=False,
            function_parameter=True,
            getter=_combine_costs_function_getter
        )
        _validate_intensity_cost_function = get_validator_by_function(is_function_type)
        _validate_adjustment_cost_function = get_validator_by_function(is_function_type)
        _validate_duration_cost_function = get_validator_by_function(is_function_type)

        # below cannot validate because the default value is None and this is considered
        # invalid. Is it that tc.typecheck only runs if an argument is specified at
        # construction?
        # _validate_modulation = get_validator_by_function(_is_modulation_param)

        def _validate_allocation_samples(self, allocation_samples):
            try:
                samples_as_array = convert_all_elements_to_np_array(allocation_samples)
                first_item_type = type(allocation_samples[0])
                first_item_shape = samples_as_array[0].shape

                for i in range(1, len(allocation_samples)):
                    if not isinstance(allocation_samples[i], first_item_type):
                        return 'all items must have the same type'
                    if not samples_as_array[i].shape == first_item_shape:
                        return 'all items must have the same shape'
            except (TypeError, IndexError):
                # not iterable, so assume single value
                pass

    portAttributes = ModulatorySignal.portAttributes | {ALLOCATION_SAMPLES,
                                                          COST_OPTIONS,
                                                          INTENSITY_COST_FUNCTION,
                                                          ADJUSTMENT_COST_FUNCTION,
                                                          DURATION_COST_FUNCTION,
                                                          COMBINE_COSTS_FUNCTION}

    connectsWith = [PARAMETER_PORT, INPUT_PORT, OUTPUT_PORT]
    connectsWithAttribute = [PARAMETER_PORTS, INPUT_PORTS, OUTPUT_PORTS]
    projectionSocket = RECEIVER
    modulators = []
    projection_type = CONTROL_PROJECTION

    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'OutputPortCustomClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    #endregion

    @tc.typecheck
    def __init__(self,
                 owner=None,
                 reference_value=None,
                 default_allocation=None,
                 size=None,
                 transfer_function=None,
                 cost_options:tc.optional(tc.any(CostFunctions, list))=None,
                 intensity_cost_function:tc.optional(is_function_type)=None,
                 adjustment_cost_function:tc.optional(is_function_type)=None,
                 duration_cost_function:tc.optional(is_function_type)=None,
                 combine_costs_function:tc.optional(is_function_type)=None,
                 allocation_samples=None,
                 modulation:tc.optional(str)=None,
                 modulates=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 **kwargs):

        try:
            if kwargs[FUNCTION] is not None:
                raise TypeError(
                    f'{self.__class__.__name__} automatically creates a '
                    'TransferWithCosts function, and does not accept override. '
                    'TransferWithCosts uses the transfer_function parameter.'
                )
        except KeyError:
            pass

        # This is included in case ControlSignal was created by another Component (such as ControlProjection)
        #    that specified ALLOCATION_SAMPLES in params
        if params and ALLOCATION_SAMPLES in params and params[ALLOCATION_SAMPLES] is not None:
            allocation_samples = params[ALLOCATION_SAMPLES]

        # FIX: ??MOVE THIS TO _validate_params OR ANOTHER _instantiate METHOD??
        # IMPLEMENTATION NOTE:
        # Consider adding self to owner.output_ports here (and removing from ControlProjection._instantiate_sender)
        #  (test for it, and create if necessary, as per OutputPorts in ControlProjection._instantiate_sender),

        # Validate sender (as variable) and params
        super().__init__(
            owner=owner,
            reference_value=reference_value,
            default_allocation=default_allocation,
            size=size,
            transfer_function=transfer_function,
            modulation=modulation,
            modulates=modulates,
            cost_options=cost_options,
            intensity_cost_function=intensity_cost_function,
            adjustment_cost_function=adjustment_cost_function,
            duration_cost_function=duration_cost_function,
            combine_costs_function=combine_costs_function,
            allocation_samples=allocation_samples,
            params=params,
            name=name,
            prefs=prefs,
            **kwargs
        )

    def _validate_params(self, request_set, target_set=None, context=None):
        """Validate cost functions and allocation_samples

        Checks if:
        - cost functions are all appropriate
           (i.e., are references to valid ControlProjection costFunctions (listed in self.costFunctions)
        - allocation_samples is a list, array, range or SampleSpec

        """

        # FIX: REINSTATE IF WANT VALIDATION ON CONTROLSIGNAL RATHER THAN ITS FUNCTION (TO HIDE FROM USER)
        # # MODIFIED 8/30/19 OLD:
        # # Validate cost functions in request_set
        # #   This should be all of them if this is an initialization call;
        # for cost_function_name in [item for item in request_set if item in costFunctionNames]:
        #     cost_function = request_set[cost_function_name]
        #
        #     # cost function assigned None: OK
        #     if not cost_function:
        #         continue
        #
        #     # cost_function is Function class specification:
        #     #    instantiate it and test below
        #     if inspect.isclass(cost_function) and issubclass(cost_function, Function):
        #         cost_function = cost_function()
        #
        #     # cost_function is Function object:
        #     #     COMBINE_COSTS_FUNCTION must be CombinationFunction
        #     #     DURATION_COST_FUNCTION must be an IntegratorFunction
        #     #     others must be TransferFunction
        #     if isinstance(cost_function, Function):
        #         if cost_function_name == COMBINE_COSTS_FUNCTION:
        #             if not isinstance(cost_function, CombinationFunction):
        #                 raise ControlSignalError("Assignment of Function to {} ({}) must be a CombinationFunction".
        #                                          format(COMBINE_COSTS_FUNCTION, cost_function))
        #         elif cost_function_name == DURATION_COST_FUNCTION:
        #             if not isinstance(cost_function, IntegratorFunction):
        #                 raise ControlSignalError("Assignment of Function to {} ({}) must be an IntegratorFunction".
        #                                          format(DURATION_COST_FUNCTION, cost_function))
        #         elif not isinstance(cost_function, TransferFunction):
        #             raise ControlSignalError("Assignment of Function to {} ({}) must be a TransferFunction".
        #                                      format(cost_function_name, cost_function))
        #
        #     # cost_function is custom-specified function
        #     #     DURATION_COST_FUNCTION and COMBINE_COSTS_FUNCTION must accept an array
        #     #     others must accept a scalar
        #     #     all must return a scalar
        #     elif isinstance(cost_function, (types.FunctionType, types.MethodType)):
        #         if cost_function_name in COMBINE_COSTS_FUNCTION:
        #             test_value = [1, 1]
        #         else:
        #             test_value = 1
        #         try:
        #             result = cost_function(test_value)
        #             if not (is_numeric(result) or is_numeric(np.asscalar(result))):
        #                 raise ControlSignalError("Function assigned to {} ({}) must return a scalar".
        #                                          format(cost_function_name, cost_function))
        #         except:
        #             raise ControlSignalError("Function assigned to {} ({}) must accept {}".
        #                                      format(cost_function_name, cost_function, type(test_value)))
        #
        #     # Unrecognized function assignment
        #     else:
        #         raise ControlSignalError("Unrecognized function ({}) assigned to {}".
        #                                  format(cost_function, cost_function_name))
        # MODIFIED 8/30/19 END

        # Validate allocation samples list:
        # - default is 1D np.array (defined by self.class_defaults.allocation_samples)
        # - however, for convenience and compatibility, allow lists:
        #    check if it is a list of numbers, and if so convert to np.array
        if ALLOCATION_SAMPLES in request_set:
            from psyneulink.core.globals.sampleiterator import allowable_specs
            allocation_samples = request_set[ALLOCATION_SAMPLES]
            if isinstance(allocation_samples, list):
                if iscompatible(allocation_samples, **{kwCompatibilityType: list,
                                                           kwCompatibilityNumeric: True,
                                                           kwCompatibilityLength: False,
                                                           }):
                    # Convert to np.array to be compatible with default value
                    request_set[ALLOCATION_SAMPLES] = np.array(allocation_samples)
            elif isinstance(allocation_samples, np.ndarray) and allocation_samples.ndim == 1:
                pass
            # elif all(isinstance(allocation_samples, spec) for spec in allowable_specs):
            elif is_sample_spec(allocation_samples):
                pass
            else:
                raise ControlSignalError("allocation_samples argument ({}) in {} must be "
                                         "a list or 1D array of numbers, a range, or a {}".
                                         format(allocation_samples, self.name, SampleSpec.__name__))

        super()._validate_params(request_set=request_set, target_set=target_set, context=context)

        # FIX: ??REINSTATE IF ABOVE IS REINSTATED??
        # # MODIFIED 8/30/19 OLD:
        # # ControlProjection Cost Functions
        # for cost_function_name in [item for item in target_set if item in costFunctionNames]:
        #     cost_function = target_set[cost_function_name]
        #     if not cost_function:
        #         continue
        #     if ((not isinstance(cost_function, (Function, types.FunctionType, types.MethodType)) and
        #              not issubclass(cost_function, Function))):
        #         raise ControlSignalError("{0} not a valid Function".format(cost_function))
        # MODIFIED 8/30/19 END

    def _instantiate_attributes_before_function(self, function=None, context=None):

        super()._instantiate_attributes_before_function(function=function, context=context)
        self._instantiate_allocation_samples(context=context)

    def _instantiate_function(self, function, function_params=None, context=None):

        from psyneulink.core.components.functions.transferfunctions import \
            TRANSFER_FCT, INTENSITY_COST_FCT, ADJUSTMENT_COST_FCT, DURATION_COST_FCT, COMBINE_COSTS_FCT

        fcts = {
            TRANSFER_FCT:self.defaults.transfer_function,
            INTENSITY_COST_FCT:self.defaults.intensity_cost_function,
            ADJUSTMENT_COST_FCT:self.defaults.adjustment_cost_function,
            DURATION_COST_FCT:self.defaults.duration_cost_function,
            COMBINE_COSTS_FCT:self.defaults.combine_costs_function,
        }
        function = TransferWithCosts(default_variable=self.defaults.variable,
                                     enabled_cost_functions=self.defaults.cost_options,
                                     **fcts)

        super()._instantiate_function(function, function_params, context)

    def _instantiate_allocation_samples(self, context=None):
        """Assign specified `allocation_samples <ControlSignal.allocation_samples>` to a `SampleIterator`."""

        a = self.parameters.allocation_samples._get(context)

        if a is None:
            return

        if isinstance(a, (range, np.ndarray)):
            a = list(a)
        self.parameters.allocation_samples._set(SampleIterator(specification=a), context)

    def _instantiate_cost_attributes(self, context=None):

        if self.cost_options:
            # Default cost params
            if self.initialization_status != ContextFlags.DEFERRED_INIT:
                self.intensity_cost = self.intensity_cost_function(self.defaults.allocation)
            else:
                self.intensity_cost = self.intensity_cost_function(self.class_defaults.allocation)
            self.defaults.intensity_cost = self.intensity_cost
            self.adjustment_cost = 0
            self.duration_cost = 0
            self.cost = self.defaults.cost = self.intensity_cost

    def _parse_port_specific_specs(self, owner, port_dict, port_specific_spec):
        """Get ControlSignal specified for a parameter or in a 'control_signals' argument

        Tuple specification can be:
            (parameter name, Mechanism)
            [TBI:] (Mechanism, parameter name, weight, exponent, projection_specs)

        Returns params dict with CONNECTIONS entries if any of these was specified.

        """
        from psyneulink.core.components.projections.projection import _parse_connection_specs
        from psyneulink.core.globals.keywords import PROJECTIONS

        params_dict = {}
        port_spec = port_specific_spec

        if isinstance(port_specific_spec, dict):
            return None, port_specific_spec

        elif isinstance(port_specific_spec, tuple):

            port_spec = None
            params_dict[PROJECTIONS] = _parse_connection_specs(connectee_port_type=self,
                                                               owner=owner,
                                                               connections=port_specific_spec)
        elif port_specific_spec is not None:
            raise ControlSignalError("PROGRAM ERROR: Expected tuple or dict for {}-specific params but, got: {}".
                                  format(self.__class__.__name__, port_specific_spec))

        if params_dict[PROJECTIONS] is None:
            raise ControlSignalError("PROGRAM ERROR: No entry found in {} params dict for {} "
                                     "with specification of parameter's Mechanism or ControlProjection(s) to it".
                                        format(CONTROL_SIGNAL, owner.name))

        return port_spec, params_dict

    def _update(self, params=None, context=None):
        """Update value (intensity) and costs
        """
        super()._update(params=params, context=context)

        if self.parameters.cost_options._get(context):
            intensity = self.parameters.value._get(context)
            self.parameters.cost._set(self.compute_costs(intensity, context), context)

    def compute_costs(self, intensity, context=None):
        """Compute costs based on self.value (`intensity <ControlSignal.intensity>`)."""
        # FIX 8/30/19: NEED TO DEAL WITH DURATION_COST AS STATEFUL:  DON'T WANT TO MESS UP MAIN VALUE

        cost_options = self.parameters.cost_options._get(context)

        # try:
        #     intensity_change = intensity - self.parameters.intensity.get_previous(context)
        # except TypeError:
        #     intensity_change = [0]

        # COMPUTE COST(S)
        # Initialize as backups for cost function that are not enabled
        intensity_cost = adjustment_cost = duration_cost = 0

        if CostFunctions.INTENSITY & cost_options:
            intensity_cost = self.intensity_cost_function(intensity, context)
            self.parameters.intensity_cost._set(intensity_cost, context)

        if CostFunctions.ADJUSTMENT & cost_options:
            try:
                intensity_change = intensity - self.parameters.intensity.get_previous(context)
            except TypeError:
                intensity_change = [0]
            adjustment_cost = self.adjustment_cost_function(intensity_change, context)
            self.parameters.adjustment_cost._set(adjustment_cost, context)

        if CostFunctions.DURATION & cost_options:
            duration_cost = self.duration_cost_function(self.parameters.cost._get(context), context=context)
            self.parameters.duration_cost._set(duration_cost, context)

        return max(0.0,
                   self.combine_costs_function([intensity_cost,
                                                adjustment_cost,
                                                duration_cost],
                                               context=context))



from . import integrator
from . import leabramechanism
from . import objective
from . import transfer

from .integrator import *
from .leabramechanism import *
from .objective import *
from .transfer import *

__all__ = list(integrator.__all__)
__all__.extend(leabramechanism.__all__)
__all__.extend(objective.__all__)
__all__.extend(transfer.__all__)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* LLVM IR Generation **************************************************************
import ast
import warnings


from llvmlite import ir
from contextlib import contextmanager


from psyneulink.core.globals.keywords import AFTER, BEFORE
from psyneulink.core.scheduling.condition import Never
from psyneulink.core.scheduling.time import TimeScale
from . import helpers
from .debug import debug_env


def gen_node_wrapper(ctx, composition, node, *, tags:frozenset):
    assert "node_wrapper" in tags
    func_tags = tags.difference({"node_wrapper"})

    node_function = ctx.import_llvm_function(node, tags=func_tags)
    # FIXME: This is a hack
    is_mech = hasattr(node, 'function')

    data_struct_ptr = ctx.get_data_struct_type(composition).as_pointer()
    args = [
        ctx.get_state_struct_type(composition).as_pointer(),
        ctx.get_param_struct_type(composition).as_pointer(),
        ctx.get_input_struct_type(composition).as_pointer(),
        data_struct_ptr, data_struct_ptr]

    if not is_mech and "reset" not in tags:
        # Add condition struct of the parent composition
        # This includes structures of all nested compositions
        cond_gen = helpers.ConditionGenerator(ctx, composition)
        cond_ty = cond_gen.get_condition_struct_type().as_pointer()
        args.append(cond_ty)

    builder = ctx.create_llvm_function(args, node, "comp_wrap_" + node_function.name)
    llvm_func = builder.function
    for a in llvm_func.args:
        a.attributes.add('nonnull')

    state, params, comp_in, data_in, data_out = llvm_func.args[:5]

    if node is composition.input_CIM:
        # if there are incoming modulatory projections,
        # the input structure is shared
        if composition.parameter_CIM.afferents:
            node_in = builder.gep(comp_in, [ctx.int32_ty(0), ctx.int32_ty(0)])
        else:
            node_in = comp_in
        incoming_projections = []
    elif node is composition.parameter_CIM and node.afferents:
        # if parameter_CIM has afferent projections,
        # their values are in comp_in[1]
        node_in = builder.gep(comp_in, [ctx.int32_ty(0), ctx.int32_ty(1)])
        # And we run no further projection
        incoming_projections = []
    elif not is_mech:
        node_in = builder.alloca(node_function.args[2].type.pointee)
        incoming_projections = node.input_CIM.afferents + node.parameter_CIM.afferents
    else:
        # this path also handles parameter_CIM with no afferent
        # projections. 'comp_in' does not include any extra values,
        # and the entire call should be optimized out.
        node_in = builder.alloca(node_function.args[2].type.pointee)
        incoming_projections = node.afferents

    if "reset" in tags:
        incoming_projections = []

    # Execute all incoming projections
    inner_projections = list(composition._inner_projections)
    zero = ctx.int32_ty(0)
    projections_params = helpers.get_param_ptr(builder, composition,
                                               params, "projections")
    projections_states = helpers.get_state_ptr(builder, composition,
                                               state, "projections")
    for proj in incoming_projections:
        # Skip autoassociative projections.
        # Recurrent projections are executed as part of the mechanism to
        # make sure their value is up-to-date inside the 'is_finished' loop
        if proj.sender.owner is proj.receiver.owner:
            continue

        # Get location of projection input data
        par_mech = proj.sender.owner
        if par_mech in composition._all_nodes:
            parent_idx = composition._get_node_index(par_mech)
        else:
            assert par_mech is par_mech.composition.output_CIM
            parent_idx = composition.nodes.index(par_mech.composition)

        assert proj.sender in par_mech.output_ports
        output_port_idx = par_mech.output_ports.index(proj.sender)
        proj_in = builder.gep(data_in, [ctx.int32_ty(0),
                                        ctx.int32_ty(0),
                                        ctx.int32_ty(parent_idx),
                                        ctx.int32_ty(output_port_idx)])

        # Get location of projection output (in mechanism's input structure
        rec_port = proj.receiver
        assert rec_port.owner is node or rec_port.owner is node.input_CIM or rec_port.owner is node.parameter_CIM
        indices = [0]
        if proj in rec_port.owner.path_afferents:
            rec_port_idx = rec_port.owner.input_ports.index(rec_port)

            assert proj in rec_port.pathway_projections
            projection_idx = rec_port.pathway_projections.index(proj)

            if not is_mech and node.parameter_CIM.afferents:
                # If there are afferent projections to parameter_CIM
                # the input structure is split between input_CIM
                # and parameter_CIM
                if proj in node.parameter_CIM.afferents:
                    # modulatory projection
                    indices.append(1)
                else:
                    # pathway projection
                    indices.append(0)
            indices.extend([rec_port_idx, projection_idx])
        elif proj in rec_port.owner.mod_afferents:
            # Only mechanism ports list mod projections in mod_afferents
            assert is_mech
            projection_idx = rec_port.owner.mod_afferents.index(proj)
            indices.extend([len(rec_port.owner.input_ports), projection_idx])
        else:
            assert False, "Projection neither pathway nor modulatory"

        proj_out = builder.gep(node_in, [ctx.int32_ty(i) for i in indices])

        # Get projection parameters and state
        proj_idx = inner_projections.index(proj)
        proj_params = builder.gep(projections_params, [zero, ctx.int32_ty(proj_idx)])
        proj_state = builder.gep(projections_states, [zero, ctx.int32_ty(proj_idx)])
        proj_function = ctx.import_llvm_function(proj, tags=func_tags)

        if proj_out.type != proj_function.args[3].type:
            warnings.warn("Shape mismatch: Projection ({}) results does not match the receiver state({}) input: {} vs. {}".format(proj, proj.receiver, proj.defaults.value, proj.receiver.defaults.variable))
            proj_out = builder.bitcast(proj_out, proj_function.args[3].type)
        builder.call(proj_function, [proj_params, proj_state, proj_in, proj_out])


    node_idx = ctx.int32_ty(composition._get_node_index(node))
    nodes_params = helpers.get_param_ptr(builder, composition, params, "nodes")
    nodes_states = helpers.get_state_ptr(builder, composition, state, "nodes")
    node_params = builder.gep(nodes_params, [zero, node_idx])
    node_state = builder.gep(nodes_states, [zero, node_idx])
    node_out = builder.gep(data_out, [zero, zero, node_idx])
    if is_mech:
        call_args = [node_params, node_state, node_in, node_out]
        if len(node_function.args) > 4:
            assert node is composition.controller
            call_args += [params, state, data_in]
        builder.call(node_function, call_args)
    elif "reset" not in tags:
        # FIXME: reinitialization of compositions is not supported
        # Condition and data structures includes parent first
        nested_idx = ctx.int32_ty(composition._get_node_index(node) + 1)
        node_data = builder.gep(data_in, [zero, nested_idx])
        node_cond = builder.gep(llvm_func.args[5], [zero, nested_idx])
        builder.call(node_function, [node_state, node_params, node_in,
                                     node_data, node_cond])
        # Copy output of the nested composition to its output place
        output_idx = node._get_node_index(node.output_CIM)
        result = builder.gep(node_data, [zero, zero, ctx.int32_ty(output_idx)])
        builder.store(builder.load(result), node_out)

    builder.ret_void()

    return llvm_func


@contextmanager
def _gen_composition_exec_context(ctx, composition, *, tags:frozenset, suffix="", extra_args=[]):
    cond_gen = helpers.ConditionGenerator(ctx, composition)

    name = "_".join(("wrap_exec", *tags ,composition.name + suffix))
    args = [ctx.get_state_struct_type(composition).as_pointer(),
            ctx.get_param_struct_type(composition).as_pointer(),
            ctx.get_input_struct_type(composition).as_pointer(),
            ctx.get_data_struct_type(composition).as_pointer(),
            cond_gen.get_condition_struct_type().as_pointer()]
    builder = ctx.create_llvm_function(args + extra_args, composition, name)
    llvm_func = builder.function

    for a in llvm_func.args:
        a.attributes.add('noalias')

    state, params, comp_in, data_arg, cond, *_ = llvm_func.args
    if "const_params" in debug_env:
        const_params = params.type.pointee(composition._get_param_initializer(None))
        params = builder.alloca(const_params.type, name="const_params_loc")
        builder.store(const_params, params)

    if "alloca_data" in debug_env:
        data = builder.alloca(data_arg.type.pointee)
        data_vals = builder.load(data_arg)
        builder.store(data_vals, data)
    else:
        data = data_arg

    node_tags = tags.union({"node_wrapper"})
    # Call input CIM
    input_cim_w = ctx.get_node_wrapper(composition, composition.input_CIM)
    input_cim_f = ctx.import_llvm_function(input_cim_w, tags=node_tags)
    builder.call(input_cim_f, [state, params, comp_in, data, data])

    # Call parameter CIM
    param_cim_w = ctx.get_node_wrapper(composition, composition.parameter_CIM)
    param_cim_f = ctx.import_llvm_function(param_cim_w, tags=node_tags)
    builder.call(param_cim_f, [state, params, comp_in, data, data])

    yield builder, data, params, cond_gen

    if "alloca_data" in debug_env:
        data_vals = builder.load(data)
        builder.store(data_vals, data_arg)

    # Bump run counter
    cond_gen.bump_ts(builder, cond, (1, 0, 0))

    builder.ret_void()


def gen_composition_exec(ctx, composition, *, tags:frozenset):
    simulation = "simulation" in tags
    node_tags = tags.union({"node_wrapper"})
    extra_args = []

    with _gen_composition_exec_context(ctx, composition, tags=tags, extra_args=extra_args) as (builder, data, params, cond_gen):
        state, _, comp_in, _, cond = builder.function.args

        # Reset internal clocks of each node
        for idx, node in enumerate(composition._all_nodes):
            #FIXME: This skips nested nodes
            from psyneulink import Composition
            if isinstance(node, Composition):
                continue
            node_state = builder.gep(state, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(idx)])
            num_executions_ptr = helpers.get_state_ptr(builder, node, node_state, "num_executions")
            num_exec_time_ptr = builder.gep(num_executions_ptr, [ctx.int32_ty(0), ctx.int32_ty(TimeScale.TRIAL.value)])
            builder.store(ctx.int32_ty(0), num_exec_time_ptr)

        # Check if there's anything to reset
        for node in composition._all_nodes:
            when = getattr(node, "reset_stateful_function_when", Never())
            # FIXME: This should not be necessary. The code gets DCE'd,
            # but there are still some problems with generation
            # 'reset' function
            if node is composition.controller:
                continue

            reinit_cond = cond_gen.generate_sched_condition(
                builder, when, cond, node, None)
            with builder.if_then(reinit_cond):
                node_w = ctx.get_node_wrapper(composition, node)
                node_reinit_f = ctx.import_llvm_function(node_w, tags=node_tags.union({"reset"}))
                builder.call(node_reinit_f, [state, params, comp_in, data, data])

        if simulation is False and composition.enable_controller and \
           composition.controller_mode == BEFORE:
            assert composition.controller is not None
            controller_w = ctx.get_node_wrapper(composition, composition.controller)
            controller_f = ctx.import_llvm_function(controller_w,
                                                     tags=node_tags)
            builder.call(controller_f, [state, params, comp_in, data, data])


        # Allocate run set structure
        run_set_type = ir.ArrayType(ir.IntType(1), len(composition.nodes))
        run_set_ptr = builder.alloca(run_set_type, name="run_set")
        builder.store(run_set_type(None), run_set_ptr)

        # Allocate temporary output storage
        output_storage = builder.alloca(data.type.pointee, name="output_storage")

        iter_ptr = builder.alloca(ctx.int32_ty, name="iter_counter")
        builder.store(ctx.int32_ty(0), iter_ptr)

        # Generate pointers to 'is_finished_flags' locations
        is_finished_flag_locs = {}
        for idx, node in enumerate(composition.nodes):
            node_state = builder.gep(state, [ctx.int32_ty(0), ctx.int32_ty(0),
                                             ctx.int32_ty(idx)])
            is_finished_flag_ptr = helpers.get_state_ptr(builder, node,
                                                         node_state,
                                                         "is_finished_flag")
            is_finished_flag_locs[node] = is_finished_flag_ptr

        loop_condition = builder.append_basic_block(name="scheduling_loop_condition")
        builder.branch(loop_condition)

        # Generate a while not 'end condition' loop
        builder.position_at_end(loop_condition)

        run_cond = cond_gen.generate_sched_condition(
            builder, composition.termination_processing[TimeScale.TRIAL],
            cond, None, is_finished_flag_locs)
        run_cond = builder.not_(run_cond, name="not_run_cond")

        loop_body = builder.append_basic_block(name="scheduling_loop_body")
        exit_block = builder.append_basic_block(name="exit")
        builder.cbranch(run_cond, loop_body, exit_block)

        # Generate loop body
        builder.position_at_end(loop_body)

        zero = ctx.int32_ty(0)
        any_cond = ir.IntType(1)(0)
        # Calculate execution set before running the mechanisms
        for idx, node in enumerate(composition.nodes):
            run_set_node_ptr = builder.gep(run_set_ptr,
                                           [zero, ctx.int32_ty(idx)],
                                           name="run_cond_ptr_" + node.name)
            node_cond = cond_gen.generate_sched_condition(
                builder, composition._get_processing_condition_set(node),
                cond, node, is_finished_flag_locs)
            ran = cond_gen.generate_ran_this_pass(builder, cond, node)
            node_cond = builder.and_(node_cond, builder.not_(ran),
                                     name="run_cond_" + node.name)
            any_cond = builder.or_(any_cond, node_cond, name="any_ran_cond")
            builder.store(node_cond, run_set_node_ptr)

        for idx, node in enumerate(composition.nodes):
            # Reset node internal clock
            #FIXME: This skips nested nodes
            from psyneulink import Composition
            if not isinstance(node, Composition):
                node_state = builder.gep(state, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(idx)])
                num_executions_ptr = helpers.get_state_ptr(builder, node, node_state, "num_executions")
                num_exec_time_ptr = builder.gep(num_executions_ptr, [ctx.int32_ty(0), ctx.int32_ty(TimeScale.TIME_STEP.value)])
                builder.store(ctx.int32_ty(0), num_exec_time_ptr)
                # FIXME: Move pass reset to actual pass count
                num_exec_time_ptr = builder.gep(num_executions_ptr, [ctx.int32_ty(0), ctx.int32_ty(TimeScale.PASS.value)])
                builder.store(ctx.int32_ty(0), num_exec_time_ptr)

            run_set_node_ptr = builder.gep(run_set_ptr, [zero, ctx.int32_ty(idx)])
            node_cond = builder.load(run_set_node_ptr, name="node_" + node.name + "_should_run")
            with builder.if_then(node_cond):
                node_w = ctx.get_node_wrapper(composition, node)
                node_f = ctx.import_llvm_function(node_w, tags=node_tags)
                builder.block.name = "invoke_" + node_f.name
                # Wrappers do proper indexing of all structures
                # Mechanisms have only 5 args
                args = [state, params, comp_in, data, output_storage]
                if len(node_f.args) >= 6:  # Composition wrappers have 6 args
                    args.append(cond)
                builder.call(node_f, args)

                cond_gen.generate_update_after_run(builder, cond, node)
            builder.block.name = "post_invoke_" + node_f.name

        # Writeback results
        for idx, node in enumerate(composition.nodes):
            run_set_node_ptr = builder.gep(run_set_ptr, [zero, ctx.int32_ty(idx)])
            node_cond = builder.load(run_set_node_ptr, name="node_" + node.name + "_ran")
            with builder.if_then(node_cond):
                out_ptr = builder.gep(output_storage, [zero, zero, ctx.int32_ty(idx)], name="result_ptr_" + node.name)
                data_ptr = builder.gep(data, [zero, zero, ctx.int32_ty(idx)],
                                       name="data_result_" + node.name)
                builder.store(builder.load(out_ptr), data_ptr)

        # Update step counter
        with builder.if_then(any_cond):
            builder.block.name = "inc_step"
            cond_gen.bump_ts(builder, cond)

        builder.block.name = "update_iter_count"
        # Increment number of iterations
        iters = builder.load(iter_ptr, name="iterw")
        iters = builder.add(iters, ctx.int32_ty(1), name="iterw_inc")
        builder.store(iters, iter_ptr)

        max_iters = len(composition.scheduler.consideration_queue)
        completed_pass = builder.icmp_unsigned("==", iters,
                                               ctx.int32_ty(max_iters),
                                               name="completed_pass")
        # Increment pass and reset time step
        with builder.if_then(completed_pass):
            builder.block.name = "inc_pass"
            builder.store(zero, iter_ptr)
            # Bumping automatically zeros lower elements
            cond_gen.bump_ts(builder, cond, (0, 1, 0))

        builder.branch(loop_condition)

        builder.position_at_end(exit_block)

        if simulation is False and composition.enable_controller and \
           composition.controller_mode == AFTER:
            assert composition.controller is not None
            controller_w = ctx.get_node_wrapper(composition, composition.controller)
            controller_f = ctx.import_llvm_function(controller_w, tags=node_tags)
            builder.call(controller_f, [state, params, comp_in, data, data])

        # Call output CIM
        output_cim_w = ctx.get_node_wrapper(composition, composition.output_CIM)
        output_cim_f = ctx.import_llvm_function(output_cim_w, tags=node_tags)
        builder.block.name = "invoke_" + output_cim_f.name
        builder.call(output_cim_f, [state, params, comp_in, data, data])

    return builder.function


def gen_composition_run(ctx, composition, *, tags:frozenset):
    assert "run" in tags
    simulation = "simulation" in tags
    name = "_".join(("wrap",  *tags, composition.name))
    args = [ctx.get_state_struct_type(composition).as_pointer(),
            ctx.get_param_struct_type(composition).as_pointer(),
            ctx.get_data_struct_type(composition).as_pointer(),
            ctx.get_input_struct_type(composition).as_pointer(),
            ctx.get_output_struct_type(composition).as_pointer(),
            ctx.int32_ty.as_pointer(),
            ctx.int32_ty.as_pointer()]
    builder = ctx.create_llvm_function(args, composition, name)
    llvm_func = builder.function
    for a in llvm_func.args:
        a.attributes.add('noalias')

    state, params, data, data_in, data_out, runs_ptr, inputs_ptr = llvm_func.args
    # simulation does not care about the output
    # it extracts results of the controller objective mechanism
    if simulation:
        data_out.attributes.remove('nonnull')

    if not simulation and "const_data" in debug_env:
        const_data = data.type.pointee(composition._get_data_initializer(None))
        data = builder.alloca(data.type.pointee)
        builder.store(const_data, data)

    # Hardcode stateful parameters if set in the environment
    if not simulation and "const_state" in debug_env:
        const_state = state.type.pointee(composition._get_state_initializer(None))
        state = builder.alloca(const_state.type, name="const_state_loc")
        builder.store(const_state, state)

    if not simulation and "const_input" in debug_env:
        if not debug_env["const_input"]:
            input_init = [[os.defaults.variable.tolist()] for os in composition.input_CIM.input_ports]
            print("Setting default input: ", input_init)
        else:
            input_init = ast.literal_eval(debug_env["const_input"])
            print("Setting user input: ", input_init)

        builder.store(data_in.type.pointee(input_init), data_in)
        builder.store(inputs_ptr.type.pointee(1), inputs_ptr)

    # Allocate and initialize condition structure
    cond_gen = helpers.ConditionGenerator(ctx, composition)
    cond_type = cond_gen.get_condition_struct_type()
    cond = builder.alloca(cond_type)
    cond_init = cond_type(cond_gen.get_condition_initializer())
    builder.store(cond_init, cond)

    runs = builder.load(runs_ptr, "runs")
    with helpers.for_loop_zero_inc(builder, runs, "run_loop") as (b, iters):
        # Get the right input stimulus
        input_idx = b.urem(iters, b.load(inputs_ptr))
        data_in_ptr = b.gep(data_in, [input_idx])

        # Reset internal clocks of each node
        for idx, node in enumerate(composition._all_nodes):
            #FIXME: This skips nested nodes
            from psyneulink import Composition
            if isinstance(node, Composition):
                continue
            node_state = builder.gep(state, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(idx)])
            num_executions_ptr = helpers.get_state_ptr(builder, node, node_state, "num_executions")
            num_exec_time_ptr = builder.gep(num_executions_ptr, [ctx.int32_ty(0), ctx.int32_ty(TimeScale.RUN.value)])
            builder.store(ctx.int32_ty(0), num_exec_time_ptr)

        # Call execution
        exec_tags = tags.difference({"run"})
        exec_f = ctx.import_llvm_function(composition, tags=exec_tags)
        b.call(exec_f, [state, params, data_in_ptr, data, cond])

        if not simulation:
            # Extract output_CIM result
            idx = composition._get_node_index(composition.output_CIM)
            result_ptr = b.gep(data, [ctx.int32_ty(0), ctx.int32_ty(0),
                                      ctx.int32_ty(idx)])
            output_ptr = b.gep(data_out, [iters])
            result = b.load(result_ptr)
            b.store(result, output_ptr)

    builder.ret_void()
    return llvm_func


def gen_multirun_wrapper(ctx, function: ir.Function) -> ir.Function:
    if function.module is not ctx.module:
        function = ir.Function(ctx.module, function.type.pointee, function.name)
        assert function.is_declaration

    args = [a.type for a in function.args]
    args.append(ctx.int32_ty.as_pointer())
    multirun_ty = ir.FunctionType(function.type.pointee.return_type, args)
    multirun_f = ir.Function(ctx.module, multirun_ty, function.name + "_multirun")
    block = multirun_f.append_basic_block(name="entry")
    builder = ir.IRBuilder(block)

    multi_runs = builder.load(multirun_f.args[-1])
    # Runs need special handling. data_in and data_out are one dimensional,
    # but hold entries for all parallel invocations.
    is_comp_run = len(function.args) == 7
    if is_comp_run:
        runs_count = builder.load(multirun_f.args[5])
        input_count = builder.load(multirun_f.args[6])

    with helpers.for_loop_zero_inc(builder, multi_runs, "multi_run_loop") as (b, index):
        # Index all pointer arguments
        indexed_args = []
        for i, arg in enumerate(multirun_f.args[:-1]):
            # Don't adjust #inputs and #trials
            if isinstance(arg.type, ir.PointerType):
                offset = index
                # #runs and #trials needs to be the same for every invocation
                if is_comp_run and i >= 5:
                    offset = ctx.int32_ty(0)
                # data arrays need special handling
                elif is_comp_run and i == 4:  # data_out
                    offset = b.mul(index, runs_count)
                elif is_comp_run and i == 3:  # data_in
                    offset = b.mul(index, input_count)

                arg = b.gep(arg, [offset])

            indexed_args.append(arg)

        b.call(function, indexed_args)

    builder.ret_void()
    return multirun_f


def gen_autodiffcomp_exec(ctx, composition, *, tags:frozenset):
    """Creates llvm bin execute for autodiffcomp"""
    assert composition.controller is None
    composition._build_pytorch_representation(composition.default_execution_id)
    pytorch_model = composition.parameters.pytorch_representation.get(composition.default_execution_id)
    with _gen_composition_exec_context(ctx, composition, tags=tags) as (builder, data, params, cond_gen):
        state, _, comp_in, _, cond = builder.function.args

        pytorch_func = ctx.import_llvm_function(pytorch_model, tags=tags)
        builder.call(pytorch_func, [state, params, data])

        node_tags = tags.union({"node_wrapper"})
        # Call output CIM
        output_cim_w = ctx.get_node_wrapper(composition, composition.output_CIM)
        output_cim_f = ctx.import_llvm_function(output_cim_w, tags=node_tags)
        builder.call(output_cim_f, [state, params, comp_in, data, data])

        return builder.function



##################################################################################################
#
# network_utils.py contains utility functions for the agent model -- specifically, a neural network
# instantiated with PsyNeuLink.
#
# Author: Yotam Sagiv
#
##################################################################################################

import numpy as np
import psyneulink as pnl
import networkx as nx
import sys
import warnings
warnings.filterwarnings("error", category=UserWarning)

###################### Convenience functions for testing script #################################

# read in bipartite graph, return graph object, number of possible tasks, number of 
# input dimensions and number of output dimensions.
# file format Ni No (input dimension number, output dimension number)
def read_bipartite_adjlist(filename):
	g = nx.Graph()
	with open(filename) as f:
		for line in f:
			inode, onode = line.split()
			g.add_node(inode, bipartite=1)
			g.add_node(onode, bipartite=0)
			g.add_edge(inode, onode, task_id=(inode, onode))

	onodes = { n for n, d in g.nodes(data=True) if d['bipartite'] == 0 }
	inodes = set(g) - onodes

	return g, g.number_of_edges(), len(inodes), len(onodes)

############################## Network utility functions #######################################

# Given LCA parameters and environment spec, return the LCA connectivity matrix
def get_LCA_matrix(num_output_dims, num_features, self_excitation, competition):
	output_layer_size = num_output_dims * num_features

	# Set up unit block matrix
	unit_dim_mat = np.zeros((num_features, num_features)) - competition
	np.fill_diagonal(unit_dim_mat, self_excitation)

	# Build overall matrix in terms of unit block matrices on the diagonal, zeros elsewhere
	for output_dim in range(num_output_dims):
		if output_dim == 0:
			lca_mat = unit_dim_mat
		else:
			lca_mat = np.block([
				[lca_mat, np.zeros((lca_mat.shape[0], unit_dim_mat.shape[1]))], 
				[np.zeros((unit_dim_mat.shape[0], lca_mat.shape[1])), unit_dim_mat],
			])

	return lca_mat

# Given a bipartite graph object, parse out a list of all the tasks defined on
# that graph.
def get_all_tasks(env_bipartite_graph):
	graph_edges = env_bipartite_graph.edges()
	all_tasks = []

	# Check if the nodes are the right way around (input, output)
	for edge in graph_edges:
		if edge[1][-1] == 'i' or edge[0][-1] == 'o': # hopefully equivalent
			inode = edge[1]
			onode = edge[0]
		else:
			inode = edge[0]
			onode = edge[1]

		# Strip suffix and convert to ints
		all_tasks.append((int(inode[:-1]), int(onode[:-1])))

	return all_tasks

# Train a multitasking network using PsyNeuLink and return the trained network, with optional attached
# LCAMechanism at the end for performance evaluation
# Params:
#     bipartite_graph: bipartite graph representing the task environment (NetworkX object)
#     num_features: number of particular features per dimension (e.g. number of colours)
#     num_hidden: number of hidden units in the network
#     epochs: number of training iterations
#     learning_rate: learning rate for SGD or (however pnl train their networks)
#     attach_LCA: if True, will attach an LCAMechanism to evaluate network performance
#     rest: LCA parameters
def get_trained_network(bipartite_graph, num_features=3, num_hidden=200, epochs=10, learning_rate=20, attach_LCA=True, competition=0.2, self_excitation=0.2, leak=0.4, threshold=1e-4):
	# Get all tasks from bipartite graph (edges) and strip 'i/o' suffix
	all_tasks = get_all_tasks(bipartite_graph)

	# Analyze bipartite graph for network properties
	onodes = [ n for n, d in bipartite_graph.nodes(data=True) if d['bipartite'] == 0 ]
	inodes = [ n for n, d in bipartite_graph.nodes(data=True) if d['bipartite'] == 1 ]
	input_dims = len(inodes)
	output_dims = len(onodes)
	num_tasks = len(all_tasks)

	# Start building network as PsyNeuLink object
	# Layer parameters
	nh = num_hidden
	D_i = num_features * input_dims
	D_c = num_tasks
	D_h = nh
	D_o = num_features * output_dims

	# Weight matrices (defaults provided by Dillon)
	wih = np.random.rand(D_i, D_h) * 0.02 - 0.01
	wch = np.random.rand(D_c, D_h) * 0.02 - 0.01
	wco = np.random.rand(D_c, D_o) * 0.02 - 0.01
	who = np.random.rand(D_h, D_o) * 0.02 - 0.01

	# Training params (defaults provided by Dillon)
	patience = 10
	min_delt = 0.00001
	lr = learning_rate

	# Instantiate layers and projections
	il = pnl.TransferMechanism(size=D_i, name='input')
	cl = pnl.TransferMechanism(size=D_c, name='control')

	hl = pnl.TransferMechanism(size=D_h, name='hidden',
	                       function=pnl.Logistic(bias=-2))

	ol = pnl.TransferMechanism(size=D_o, name='output',
	                       function=pnl.Logistic(bias=-2))

	pih = pnl.MappingProjection(matrix=wih)
	pch = pnl.MappingProjection(matrix=wch)
	pco = pnl.MappingProjection(matrix=wco)
	pho = pnl.MappingProjection(matrix=who)

	# Create training data for network
	# We train across all possible inputs, one task at a time
	input_examples, output_examples, control_examples = generate_training_data(all_tasks, num_features, input_dims, output_dims)

	# Training parameter set
	input_set = {
	    'inputs': {
	        il: input_examples.tolist(),
	        cl: control_examples.tolist()
	    },
	    'targets': {
	        ol: output_examples.tolist()
	    },
	    'epochs': epochs
	}

	# Build network
	mnet = pnl.AutodiffComposition(
                           patience=patience,
                           min_delta=min_delt,
                           learning_rate=learning_rate,
                           learning_enabled=True,
                           name='mnet')

	mnet.output_CIM.parameters.value._set_history_max_length(100000)
	mnet.add_node(il)
	mnet.add_node(cl)
	mnet.add_node(hl)
	mnet.add_node(ol)
	mnet.add_projection(projection=pih, sender=il, receiver=hl)
	mnet.add_projection(projection=pch, sender=cl, receiver=hl)
	mnet.add_projection(projection=pco, sender=cl, receiver=ol)
	mnet.add_projection(projection=pho, sender=hl, receiver=ol)

	# Train network
	mnet.run(
	    inputs=input_set,
	    minibatch_size=1,
	    bin_execute=True
	)

	for projection in mnet.projections:
		weights = projection.parameters.matrix.get(mnet)
		projection.parameters.matrix.set(weights, None)

	mnet.learning_enabled = False

	# Apply LCA transform (values from Sebastian's code -- supposedly taken from the original LCA paper from Marius & Jay)
	if attach_LCA:
		lca = pnl.LCAMechanism(size=D_o, 
							   leak=leak, 
							   competition=competition, 
							   self_excitation=self_excitation, 
							   time_step_size=0.01, 
							   threshold=threshold, 
							   threshold_criterion=pnl.CONVERGENCE,
							   reset_stateful_function_when=pnl.AtTrialStart(),
							   name='lca')

		# Wrapper composition used to pass values between mnet (AutodiffComposition) and lca (LCAMechanism)
		wrapper_composition = pnl.Composition()

		# Add mnet and lca to outer_composition
		wrapper_composition.add_linear_processing_pathway([mnet, lca])

		return wrapper_composition

	return mnet

# Train a multitasking network using PsyNeuLink and return the trained network, with optional attached
# RecurrentTransferMechanism at the end for performance evaluation
# Params:
#     bipartite_graph: bipartite graph representing the task environment (NetworkX object)
#     num_features: number of particular features per dimension (e.g. number of colours)
#     num_hidden: number of hidden units in the network
#     epochs: number of training iterations
#     learning_rate: learning rate for SGD or (however pnl train their networks)
#     attach_LCA: if True, will attach an LCAMechanism to evaluate network performance
#     rest: LCA parameters
def get_trained_network_multLCA(bipartite_graph, num_features=3, num_hidden=200, epochs=10, learning_rate=20,
								attach_LCA=True, competition=0.2, self_excitation=0.2, leak=0.4, threshold=1e-4,
								exec_limit=10000):
	# Get all tasks from bipartite graph (edges) and strip 'i/o' suffix
	all_tasks = get_all_tasks(bipartite_graph)

	# Analyze bipartite graph for network properties
	onodes = [ n for n, d in bipartite_graph.nodes(data=True) if d['bipartite'] == 0 ]
	inodes = [ n for n, d in bipartite_graph.nodes(data=True) if d['bipartite'] == 1 ]
	input_dims = len(inodes)
	output_dims = len(onodes)
	num_tasks = len(all_tasks)

	# Start building network as PsyNeuLink object
	# Layer parameters
	nh = num_hidden
	D_i = num_features * input_dims
	D_c = num_tasks
	D_h = nh
	D_o = num_features * output_dims

	# Weight matrices (defaults provided by Dillon)
	wih = np.random.rand(D_i, D_h) * 0.02 - 0.01
	wch = np.random.rand(D_c, D_h) * 0.02 - 0.01
	wco = np.random.rand(D_c, D_o) * 0.02 - 0.01
	who = np.random.rand(D_h, D_o) * 0.02 - 0.01

	# Training params (defaults provided by Dillon)
	patience = 10
	min_delt = 0.00001
	lr = learning_rate

	# Instantiate layers and projections
	il = pnl.TransferMechanism(size=D_i, name='input')
	cl = pnl.TransferMechanism(size=D_c, name='control')

	hl = pnl.TransferMechanism(size=D_h, name='hidden',
	                       function=pnl.Logistic(bias=-2))

	ol = pnl.TransferMechanism(size=D_o, name='output',
	                       function=pnl.Logistic(bias=-2))

	pih = pnl.MappingProjection(matrix=wih)
	pch = pnl.MappingProjection(matrix=wch)
	pco = pnl.MappingProjection(matrix=wco)
	pho = pnl.MappingProjection(matrix=who)

	# Create training data for network
	# We train across all possible inputs, one task at a time
	input_examples, output_examples, control_examples = generate_training_data(all_tasks, num_features, input_dims, output_dims)

	# Training parameter set
	input_set = {
	    'inputs': {
	        il: input_examples.tolist(),
	        cl: control_examples.tolist()
	    },
	    'targets': {
	        ol: output_examples.tolist()
	    },
	    'epochs': epochs
	}

	# Build network
	mnet = pnl.AutodiffComposition(
                           patience=patience,
                           min_delta=min_delt,
                           learning_rate=learning_rate,
                           learning_enabled=True,
                           name='mnet')

	mnet.output_CIM.parameters.value._set_history_max_length(1000)
	mnet.add_node(il)
	mnet.add_node(cl)
	mnet.add_node(hl)
	mnet.add_node(ol)
	mnet.add_projection(projection=pih, sender=il, receiver=hl)
	mnet.add_projection(projection=pch, sender=cl, receiver=hl)
	mnet.add_projection(projection=pco, sender=cl, receiver=ol)
	mnet.add_projection(projection=pho, sender=hl, receiver=ol)

	# Train network
	mnet.run(
	    inputs=input_set,
	    minibatch_size=1,
	    bin_execute=True
	)

	for projection in mnet.projections:
		weights = projection.parameters.matrix.get(mnet)
		projection.parameters.matrix.set(weights, None)

	mnet.learning_enabled = False

	# Apply LCA transform (values from Sebastian's code -- supposedly taken from the original LCA paper from Marius & Jay)
	if attach_LCA:
		lci = pnl.LeakyCompetingIntegrator(rate=leak, 
										   time_step_size=0.01)

		lca_matrix = get_LCA_matrix(output_dims, num_features, self_excitation, competition)

		lca = pnl.RecurrentTransferMechanism(size=D_o,
											 matrix=lca_matrix,
											 integrator_mode=True,
											 integrator_function=lci,
											 name='lca',
											 termination_threshold=threshold,
											 reset_stateful_function_when=pnl.AtTrialStart())

		# Wrapper composition used to pass values between mnet (AutodiffComposition) and lca (LCAMechanism)
		wrapper_composition = pnl.Composition()

		# Add mnet and lca to outer_composition
		wrapper_composition.add_linear_processing_pathway([mnet, lca])

		# Set execution limit
		lca.parameters.max_executions_before_finished.set(exec_limit, wrapper_composition)

		# # Logging/Debugging
		# lca.set_log_conditions('value', pnl.LogCondition.EXECUTION)

		return wrapper_composition

	return mnet


# Generate data for the network to train on. This means single-task training on all available
# tasks within the environment, under a uniform task distribution. As data we generate all possible
# mappings within each task. To specify a mapping, we use the rule that input feature nodes map to 
# equal ordinal output feature nodes (i.e. 1st feature input maps to 1st feature output).
# Params:
#     all_tasks: list containing all tasks in the environment
#     num_features: number of particular features per dimension (e.g. number of colours)
#     num_input_dims: number of input dimensions in the environment
#     num_output_dims: number of output dimensions in the environment
#     samples_per_feature: how many stimuli will be sampled per feature to be trained within a task
#               (even though the input-output feature mapping is fixed for a given task, the values
#                of all the other input dimensions are not, so we can sample many stimuli for a given
#                feature association)
def generate_training_data(all_tasks, num_features, num_input_dims, num_output_dims, samples_per_feature=100):
	# Extract relevant parameters
	num_tasks = len(all_tasks)
	input_layer_size = num_features * num_input_dims
	output_layer_size = num_features * num_output_dims
	control_layer_size = num_tasks
	num_examples = num_features * num_tasks * samples_per_feature

	# Instantiate example matrices
	input_examples = np.zeros((num_examples, input_layer_size))
	output_examples = np.zeros((num_examples, output_layer_size))
	control_examples = np.zeros((num_examples, control_layer_size))

	# Create examples, task by task
	row_count = 0
	for task in all_tasks:
		# Load parameters
		task_input_dim, task_output_dim = task
		control_idx = task_id_to_control_idx(task, all_tasks)

		# Generate feature maps (we arbitrarily pick the redundant mapping within each dimension)
		# and also randomly sample input values for the other dimensions of the stimulus
		for _ in range(samples_per_feature):
			# Set feature of relevant task
			for i in range(num_features): 
				input_examples[row_count, task_input_dim * num_features + i] = 1
				output_examples[row_count, task_output_dim * num_features + i] = 1
				control_examples[row_count, control_idx] = 1

				# Set all other stimulus dimensions randomly
				for input_dim in range(num_input_dims):
					if input_dim == task_input_dim:
						continue

					input_examples[row_count, input_dim * num_features + np.random.choice(num_features)] = 1

				row_count += 1

	return input_examples, output_examples, control_examples

# Generate data for the network to test on. test_tasks is a performance set (i.e. a multitasking set of tasks to execute). 
# As data we generate random features for all input dimensions. To specify a mapping, we use the rule that input 
# feature nodes map to equal ordinal output feature nodes (i.e. 1st feature input maps to 1st feature output).
# Params:
#     test_tasks: list containing set of tasks to multitask
#     all_tasks: list containing all tasks in the environment
#     num_features: number of particular features per dimension (e.g. number of colours)
#     num_input_dims: number of input dimensions in the environment
#     num_output_dims: number of output dimensions in the environment
#     num_test_points: number of test points to generate
def generate_testing_data(test_tasks, all_tasks, num_features, num_input_dims, num_output_dims, num_test_points):
	# Extract relevant parameters
	num_tasks = len(all_tasks)
	input_layer_size = num_features * num_input_dims
	output_layer_size = num_features * num_output_dims
	control_layer_size = num_tasks

	# Instantiate example matrices
	input_examples = np.zeros((num_test_points, input_layer_size))
	output_examples = np.zeros((num_test_points, output_layer_size))
	control_examples = np.zeros((num_test_points, control_layer_size))

	# Create examples
	for i in range(num_test_points):
		for input_dim in range(num_input_dims):
			# Input
			feature = np.random.choice(num_features)
			input_idx = input_dim * num_features + feature
			input_examples[i, input_idx] = 1

			# Output & Control
			for output_dim in range(num_output_dims):
				# If there is not a task with this input/output pair, move on
				if (input_dim, output_dim) not in test_tasks:
					continue

				# Output
				output_idx = output_dim * num_features + feature
				output_examples[i, output_idx] = 1

				# Control
				control_idx = task_id_to_control_idx((input_dim, output_dim), all_tasks)
				control_examples[i, control_idx] = 1

	return input_examples, output_examples, control_examples

# We define the control layer activation as just the index of task within the global all_tasks list.
def task_id_to_control_idx(task, all_tasks):
	return all_tasks.index(task)

# Use the LCA to evaluate network performance on a given set of tasks (performance set)
# Params:
#     mnet_lca: Multitasking network AutodiffComposition with attached LCAMechanism (PNL) 
#     test_tasks: Set of tasks to be simultaneously executed. List of tuples indicating (input dim, output dim)
#     all_tasks: Global list of all possible tasks within the environment. Formatted as test_tasks. 
#     num_features, num_output_dims, num_input_dims: ints encoding the relevant size
#     num_test_points: number of self-generated test points to test performance on
#     threshold: how many points correct to consider task correctly completed
def evaluate_net_perf_lca(mnet_lca, test_tasks, all_tasks, num_features, num_input_dims, num_output_dims, num_test_points, threshold=0.8):
	# Extract relevant parameters
	input_layer_size = num_features * num_input_dims
	output_layer_size = num_features * num_output_dims
	control_layer_size = len(all_tasks)
	num_tasks = len(test_tasks)

	# Instantiate test matrices
	input_test_pts, output_true_pts, control_test_pts = generate_testing_data(test_tasks,
																   all_tasks, 
																   num_features, 
																   num_input_dims, 
																   num_output_dims, 
																   num_test_points=num_test_points)

	# Run the outer composition, one point at a time (for debugging purposes)
	for i in range(num_test_points):
		# Construct input dict
		input_set = { 
			'inputs' : {
				mnet_lca.nodes['mnet'].nodes['input'] : input_test_pts[i, :].tolist(),
				mnet_lca.nodes['mnet'].nodes['control'] : control_test_pts[i, :].tolist()
			}
		}

		try:
			mnet_lca.run( { mnet_lca.nodes['mnet'] : input_set } )
		except Warning:
			print('input: ', input_test_pts[i, :])
			print('control: ', control_test_pts[i, :])
			print('true: ', output_true_pts[i, :])
			mnet = mnet_lca.nodes['mnet']
			lca = mnet_lca.nodes['lca']
			print('net out: ', np.array([mnet.output_CIM.parameters.value.get(mnet_lca)]).reshape(1, output_layer_size))
			print('num executions: ', mnet_lca.nodes['lca'].num_executions_before_finished)

			
			# mnet_lca.nodes['lca'].log.print_entries()
			sys.exit(0)

	# Retrieve LCA results
	lca_out = np.array(mnet_lca.parameters.results.get(mnet_lca)[-num_test_points:]).reshape(num_test_points, output_layer_size)

	# Retrieve mnet results 
	mnet = mnet_lca.nodes['mnet']

	# Brutal line of code provided by Dillon
	# Intuition: Part in list() expression is stored history. Part afterwards is the current value.
	# This distinction is important because PsyNeuLink objects update history only on next trial call.
	if num_test_points > 1:
		mnet_out = np.array(list(mnet.output_CIM.parameters.value.history[mnet_lca.default_execution_id])[
				   -num_test_points+1:] +[mnet.output_CIM.parameters.value.get(mnet_lca)]).reshape(num_test_points, output_layer_size)
	else:
		mnet_out = np.array([mnet.output_CIM.parameters.value.get(mnet_lca)]).reshape(num_test_points, output_layer_size)
	
	# Compare correctness, get a bunch of useful stats
	stats_dict = {
					'tasks' : test_tasks,
				  	'num_tasks_correct' : 0, 
				  	'inputs' : input_test_pts,
				  	'outputs' : lca_out,
				  	'true_outputs' : output_true_pts,
				  	'mnet_out' : mnet_out,
				  	'num_correct_per_task' : {},
				  	'wrong_idxs' : set()
				}

	for task in test_tasks:
		input_dim, output_dim = task

		num_test_pts_correct = 0
		for i in range(num_test_points):
			test_pt = input_test_pts[i, :]
			test_output = lca_out[i, :]

			# get input feature
			input_feature_idx = np.argmax(test_pt[input_dim * num_features : (input_dim + 1) * num_features])

			# get output feature
			output_feature_idx = np.argmax(test_output[output_dim * num_features : (output_dim + 1) * num_features])

			if input_feature_idx == output_feature_idx:
				num_test_pts_correct += 1
			else:
				stats_dict['wrong_idxs'].add(i)

		if num_test_pts_correct / num_test_points >= threshold:
			stats_dict['num_tasks_correct'] += 1

		stats_dict['num_correct_per_task'][task] = num_test_pts_correct

	return stats_dict 

# Use Mean Squared Error (MSE) to evaluate network performance without an LCA
# Params:
#     mnet_lca: Multitasking network AutodiffComposition 
#     test_tasks: Set of tasks to be simultaneously executed. List of tuples indicating (input dim, output dim)
#     all_tasks: Global list of all possible tasks within the environment. Formatted as test_tasks. 
#     num_features, num_output_dims, num_input_dims: ints encoding the relevant size
#     num_test_points: number of self-generated test points to test performance on
#     threshold: how many points correct to consider task correctly completed
def evaluate_net_perf_mse(mnet, test_tasks, all_tasks, num_features, num_input_dims, num_output_dims, num_test_points, threshold=0.8):
	# Extract relevant parameters
	input_layer_size = num_features * num_input_dims
	output_layer_size = num_features * num_output_dims
	control_layer_size = len(all_tasks)
	num_tasks = len(test_tasks)

	# Instantiate test matrices
	input_test_pts, output_true_pts, control_test_pts = generate_testing_data(test_tasks,
																   all_tasks, 
																   num_features, 
																   num_input_dims, 
																   num_output_dims, 
																   num_test_points=num_test_points)

	# Run the composition
	input_set = { 
		'inputs' : {
			mnet.nodes['input'] : input_test_pts.tolist(),
			mnet.nodes['control'] : control_test_pts.tolist()
		}
	}

	mnet.run(input_set)

	# Retrieve results
	output_test_pts = np.array(mnet.parameters.results.get(mnet)[-num_test_points:]).reshape(num_test_points, output_layer_size)

	# Analyze accuracy
	MSE = 0
	plotted_examples = 0
	for i in range(num_test_points):
		test_out = output_test_pts[i, :]
		true_out = output_true_pts[i, :]

		MSE += np.sum(np.square(test_out - true_out))

		# Plot some examples arbitrarily at random
		if np.random.uniform() < 0.05 and plotted_examples < 3 and len(test_tasks) > 1:
			print('all tasks: ', all_tasks)
			print('test tasks: ', test_tasks)
			print('input: ', input_test_pts[i, :])
			print('control: ', control_test_pts[i, :])
			print('true output: ', true_out)
			print('test output: ', test_out)
			print()
			plotted_examples += 1

	MSE /= (num_test_points * output_layer_size)

	return MSE 

############################## TESTING SCRIPT ##############################

# Trivial testing script
if __name__ == '__main__':
	np.set_printoptions(precision=7, threshold=sys.maxsize, suppress=True, linewidth=np.nan)
	verbose = False
	np.random.seed(12345)

	# Train and evaluate an mnet-LCA combo on single-tasking and multitasking

	# Params
	num_test_points = 100
	num_features = 3
	g, num_tasks, num_input_dims, num_output_dims = read_bipartite_adjlist('./data/bipartite_graphs/7-tasks.txt')
	all_tasks = get_all_tasks(g)

	# Get trained network (@ Jon: This won't work)
	mnet_lca = get_trained_network_multLCA(g, learning_rate=0.3, epochs=1000, attach_LCA = True, exec_limit=10000)

	# (@ Jon: This is the global LCA (i.e. no within dimension effects) and it does work, using LCAMechanism)
	# mnet_lca = get_trained_network(g, learning_rate=0.3, epochs=1000, attach_LCA = True)

	# Run some simulations

	# Save stats dictionaries to list
	perf_dicts = []

	# single tasking
	# for task in all_tasks:
	# 	perf_dicts.append(evaluate_net_perf_lca(mnet_lca, [task], all_tasks, num_features, num_input_dims, num_output_dims, num_test_points))
	
	# perf_dicts.append(evaluate_net_perf_lca(mnet_lca, [(4,3)], all_tasks, num_features, num_input_dims, num_output_dims, num_test_points))
	
	# # multitasking
	for task_i in all_tasks[:2]:
		for task_j in all_tasks[:2]:
			if task_i == task_j:
				continue

			perf_dicts.append(evaluate_net_perf_lca(mnet_lca, [task_i, task_j], all_tasks, num_features, num_input_dims, num_output_dims, num_test_points))
	
	# perf_dicts.append(evaluate_net_perf_lca(mnet_lca, [(0, 0), (1, 1)], all_tasks, num_features, num_input_dims, num_output_dims, num_test_points))

	# Print summary stats
	for perf_dict in perf_dicts:
		print('tasks: ', perf_dict['tasks'])
		print('num_tasks_correct: ', perf_dict['num_tasks_correct'])
		print('number of test points incorrect:', len(perf_dict['wrong_idxs']))
		print('number of test points correct per task:')
		for task in perf_dict['num_correct_per_task'].keys():
			print('\t', task, ' ', perf_dict['num_correct_per_task'][task])

		if verbose:
			print('all points:')
			for idx in range(num_test_points):
				print('\tidx ', idx)
				print('\tinput: ', perf_dict['inputs'][idx, :])
				print('\ttrue out: ', perf_dict['true_outputs'][idx, :])
				print('\tnet out: ', perf_dict['mnet_out'][idx, :])
				print('\tlca out: ', perf_dict['outputs'][idx, :])
				print()


		print('*****')

	# MSE Testing
	# np.random.seed(2)
	# g, num_tasks, num_input_dims, num_output_dims = read_bipartite_adjlist('./data/bipartite_graphs/7-tasks.txt')

	# mnet = get_trained_network(g, learning_rate=0.3, epochs=5000, attach_LCA = False)

	# all_tasks = get_all_tasks(g)

	# 

	# # single tasking (MSE)
	# print('******** single-tasking: ********')
	# print('%.2f' % evaluate_net_perf_mse(mnet, [(1, 1)], all_tasks, 3, num_input_dims, num_output_dims, 100))
	# print('%.2f' % evaluate_net_perf_mse(mnet, [(0, 0)], all_tasks, 3, num_input_dims, num_output_dims, 100))
	# print('%.2f' % evaluate_net_perf_mse(mnet, [(2, 3)], all_tasks, 3, num_input_dims, num_output_dims, 100))
	# print()

	# # multitasking (MSE)
	# print('******** multitasking: ********')
	# print('%.2f\n\n\n' % evaluate_net_perf_mse(mnet, [(0, 0), (1, 1)], all_tasks, 3, num_input_dims, num_output_dims, 100))
	# print('%.2f\n\n\n' % evaluate_net_perf_mse(mnet, [(0, 0), (4, 4)], all_tasks, 3, num_input_dims, num_output_dims, 100))
	# print('%.2f\n\n\n' % evaluate_net_perf_mse(mnet, [(1, 1), (2, 2)], all_tasks, 3, num_input_dims, num_output_dims, 100))

	# print('%.2f\n\n\n' % evaluate_net_perf_mse(mnet, [(1, 1), (2, 2), (3, 3), (4, 4)], all_tasks, 3, num_input_dims, num_output_dims, 100))








# -*- coding: utf-8 -*-
"""ContentAddressableMemorytesting.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1XjXuOsceGgDSyGsSF8OVQvQS4h-L-U4F
"""

# !pip install --upgrade psyneulink

import psyneulink as pnl
import numpy as np
print(pnl.__version__)

# network params
n_input = 2
n_hidden = 5
n_output = 1
max_entries = 7

# training params
num_epochs = 3
learning_rate = .1
wts_init_scale = .1

# layers

input = pnl.TransferMechanism(
    name='input',
    default_variable=np.zeros(n_input)
)

hidden = pnl.TransferMechanism(
    name='hidden',
    default_variable=np.zeros(n_hidden),
    function=pnl.Logistic()
)

output = pnl.TransferMechanism(
    name='output',
    default_variable=np.zeros(n_output),
    function=pnl.Logistic()
)

# weights
w_ih = pnl.MappingProjection(
    name='input_to_hidden',
    matrix=np.random.randn(n_input, n_hidden) * wts_init_scale,
    sender=input,
    receiver=hidden
)

w_ho = pnl.MappingProjection(
    name='hidden_to_output',
    matrix=np.random.randn(n_hidden, n_output) * wts_init_scale,
    sender=hidden,
    receiver=output
)

# ContentAddressableMemory
ContentAddressableMemory = pnl.EpisodicMemoryMechanism(
    cue_size=n_hidden, assoc_size=n_hidden,
    name='ContentAddressableMemory'
)

w_hdc = pnl.MappingProjection(
    name='hidden_to_cue',
    matrix=np.random.randn(n_hidden, n_hidden) * wts_init_scale,
    sender=hidden,
    receiver=ContentAddressableMemory.input_ports[pnl.CUE_INPUT]
)

w_hda = pnl.MappingProjection(
    name='hidden_to_assoc',
    matrix=np.random.randn(n_hidden, n_hidden) * wts_init_scale,
    sender=hidden,
    receiver=ContentAddressableMemory.input_ports[pnl.ASSOC_INPUT]
)


w_dh = pnl.MappingProjection(
    name='em_to_hidden',
    matrix=np.random.randn(n_hidden, n_hidden) * wts_init_scale,
    sender=ContentAddressableMemory,
    receiver=hidden
)

comp = pnl.Composition(name='xor')

# add all nodes
all_nodes = [input, hidden, output, ContentAddressableMemory]
for node in all_nodes:
    comp.add_node(node)
# input-hidden-output pathway
comp.add_projection(sender=input, projection=w_ih, receiver=hidden)
comp.add_projection(sender=hidden, projection=w_ho, receiver=output)
# conneciton, ContentAddressableMemory
comp.add_projection(sender=ContentAddressableMemory, projection=w_dh, receiver=hidden)
comp.add_projection(
    sender=hidden,
    projection=w_hdc,
    receiver=ContentAddressableMemory.input_ports[pnl.CUE_INPUT]
)
comp.add_projection(
    sender=hidden,
    projection=w_hda,
    receiver=ContentAddressableMemory.input_ports[pnl.ASSOC_INPUT]
)
# show graph
comp.show_graph(show_node_structure=True)

# # comp.show()
# # the required inputs for ContentAddressableMemory
# print('ContentAddressableMemory input_ports: ', ContentAddressableMemory.input_ports.names)
#
# # currently, ContentAddressableMemory receive info from the following node
# print('ContentAddressableMemory receive: ')
# for ContentAddressableMemory_input in ContentAddressableMemory.input_ports.names:
#     afferents = ContentAddressableMemory.input_ports[ContentAddressableMemory_input].path_afferents
#     if len(afferents) == 0:
#         print(f'- {ContentAddressableMemory_input}: NA')
#     else:
#         sending_node_name = afferents[0].sender.owner.name
#         print(f'- {ContentAddressableMemory_input}: {sending_node_name}')
#
# print('ContentAddressableMemory cue input: ', ContentAddressableMemory.input_ports.names)
#
# print('hidden receive: ')
# for hidden_afferent in hidden.input_ports[0].path_afferents:
#     print('- ', hidden_afferent.sender.owner.name)

# comp.show()
 # the required inputs for ContentAddressableMemory
print('ContentAddressableMemory input_ports: ', ContentAddressableMemory.input_ports.names)
  
# currently, ContentAddressableMemory receive info from the following node
print('ContentAddressableMemory receive: ')
for ContentAddressableMemory_input in ContentAddressableMemory.input_ports.names:
    afferents = ContentAddressableMemory.input_ports[ContentAddressableMemory_input].path_afferents
    if len(afferents) == 0:
        print(f'- {ContentAddressableMemory_input}: NA')
    else:
        sending_node_name = afferents[0].sender.owner.name
        print(f'- {ContentAddressableMemory_input}: {sending_node_name}')

print('ContentAddressableMemory cue input: ', ContentAddressableMemory.input_ports.names)

print('hidden receive: ')
for hidden_afferent in hidden.input_ports[0].path_afferents:
    print('- ', hidden_afferent.sender.owner.name)

print(ContentAddressableMemory.output_ports.names)
print(ContentAddressableMemory.output_ports.values)

print(input)


#comp.run([1,1])
#print(ContentAddressableMemory.values)
#ContentAddressableMemory.dict.insert_memory([1,1])

print(ContentAddressableMemory.output_ports.values)

input = [0,1]
print(ContentAddressableMemory.input_ports)
print(hidden.value)
print(output.value)


#comp.run(input)

print(comp.run([1,1]))
print(comp.run([2,2]))

print(comp.run([100,100]))

print(comp.run([10000,100000]))

#comp.output.value

input_dict = [[0,1], [1,2], [2,3], [3,4], [4,5]]
result = comp.run(inputs=input_dict, context = 5) #, do_logging=True)
print(ContentAddressableMemory.input_values)
print(hidden.value)

print(ContentAddressableMemory.output_values)

result1 = comp.run([0,1])
result2 = comp.run([0,2])

print(result1)
print(result2)

print(output.output_values)

ContentAddressableMemory.CUE_INPUT = [1,1,1,1,1]
ContentAddressableMemory.ASSOC_INPUT = [1,2,3,4,5]

print(ContentAddressableMemory.input_values)
print(ContentAddressableMemory.output_values)

# ContentAddressableMemory.function.add_to_memory([[[100,101,102,103,104],[23,24,25,26,27]]], context=5)
# assert True

# ContentAddressableMemory.dict.insert_memory({0,1})
#
# ContentAddressableMemory.get_memory(0)

"""cant figure out get&store memory functions"""



from . import statefulfunction
from . import integratorfunctions
from . import memoryfunctions

from .statefulfunction import *
from .integratorfunctions import *
from .memoryfunctions import *

__all__ = list(statefulfunction.__all__)
__all__.extend(integratorfunctions.__all__)
__all__.extend(memoryfunctions.__all__)



import numpy as np
import pytest

import psyneulink.core.components.functions.transferfunctions as Functions
import psyneulink.core.llvm as pnlvm

@pytest.mark.function
@pytest.mark.identity_function
@pytest.mark.parametrize("size", [1, 2, 4, 8, 16])
@pytest.mark.benchmark(group="IdentityFunction")
def test_basic(size, benchmark):
    variable = np.random.rand(size)
    f = Functions.Identity(default_variable=variable)
    res = benchmark(f.function, variable)
    assert np.allclose(res, variable)

@pytest.mark.llvm
@pytest.mark.function
@pytest.mark.identity_function
@pytest.mark.parametrize("size", [1, 2, 4, 8, 16])
@pytest.mark.benchmark(group="IdentityFunction")
def test_llvm(size, benchmark):
    variable = np.random.rand(size)
    f = Functions.Identity(default_variable=variable)
    m = pnlvm.execution.FuncExecution(f)
    res = benchmark(m.execute, variable)
    assert np.allclose(res, variable)

@pytest.mark.llvm
@pytest.mark.cuda
@pytest.mark.function
@pytest.mark.identity_function
@pytest.mark.parametrize("size", [1, 2, 4, 8, 16])
@pytest.mark.benchmark(group="IdentityFunction")
def test_ptx_cuda(size, benchmark):
    variable = np.random.rand(size)
    f = Functions.Identity(default_variable=variable)
    m = pnlvm.execution.FuncExecution(f)
    res = benchmark(m.cuda_execute, variable)
    assert np.allclose(res, variable)
